from typing import List, Literal, Optional

from pydantic_xml import BaseXmlModel, attr, element, wrapped


class Name(BaseXmlModel):
    language: str = attr()
    value: str = attr()
    main: bool = attr(default=False)


class Statement(BaseXmlModel):
    charset: Optional[Literal['UTF-8']] = attr(default=None)

    language: str = attr()

    mathjax: bool = attr(default=False)

    path: str = attr()

    type: Literal['application/x-tex', 'application/pdf', 'text/html'] = attr()


class File(BaseXmlModel):
    path: str = attr()
    type: Optional[str] = attr(default=None)


class Test(BaseXmlModel):
    method: Literal['manual', 'generated'] = attr()

    sample: Optional[bool] = attr(default=None)

    description: Optional[str] = attr(default=None)


class Testset(BaseXmlModel):
    name: Optional[str] = attr(default=None)

    timelimit: Optional[int] = element('time-limit', default=1000)
    memorylimit: Optional[int] = element('memory-limit', default=256 * 1024 * 1024)

    size: int = element('test-count', default=None)

    inputPattern: str = element('input-path-pattern')
    answerPattern: str = element('answer-path-pattern')

    tests: List[Test] = wrapped('tests', element(tag='test'), default_factory=list)


class Judging(BaseXmlModel):
    inputFile: str = attr(default='')
    outputFile: str = attr(default='')

    testsets: List[Testset] = element(tag='testset', default_factory=list)


class Checker(BaseXmlModel):
    name: str = attr()
    type: Literal['testlib'] = attr()
    source: File = element()
    binary: Optional[File] = element(default=None)
    cpy: Optional[File] = element(tag='copy', default=None)

    testset: Optional[Testset] = element(default=None)


class Interactor(BaseXmlModel):
    source: File = element()


class Problem(BaseXmlModel, tag='problem'):
    names: List[Name] = wrapped('names', element(tag='name'), default_factory=list)

    statements: List[Statement] = wrapped(
        'statements',
        element(tag='statement', default=[]),
        default=[],
    )

    judging: Judging = element()

    files: List[File] = wrapped(
        'files/resources',
        element(tag='file', default=[]),
        default=[],
    )

    checker: Checker = wrapped('assets', element(tag='checker'))

    interactor: Optional[Interactor] = wrapped('assets', element(tag='interactor'))


class ContestProblem(BaseXmlModel):
    index: str = attr()
    path: str = attr()


class Contest(BaseXmlModel, tag='contest'):
    names: List[Name] = wrapped('names', element(tag='name'), default_factory=list)

    statements: List[Statement] = wrapped(
        'statements',
        element(tag='statement', default=[]),
        default=[],
    )

    problems: List[ContestProblem] = wrapped(
        'problems',
        element(tag='problem', default=[]),
        default=[],
    )



import os
import pathlib
import tempfile
from typing import Any, Dict, Optional

import rich
import rich.progress
import typer

from rbx import console
from rbx.box import header, package
from rbx.box.generators import get_all_built_testcases
from rbx.box.packaging.polygon import polygon_api as api
from rbx.box.packaging.polygon.packager import code_to_langs, is_valid_lang_code
from rbx.box.schema import CodeItem, ExpectedOutcome, Solution, TaskType, Testcase
from rbx.box.statements.build_statements import get_relative_assets
from rbx.box.statements.builders import (
    StatementBlocks,
    StatementBuilderProblem,
    render_jinja_blocks,
)
from rbx.box.statements.schema import Statement, StatementType
from rbx.box.testcase_utils import get_alternate_interaction_texts, parse_interaction

_API_URL = 'https://polygon.codeforces.com/api'

POLY = api.Polygon(
    _API_URL,
    os.environ.get('POLYGON_API_KEY', '').strip(),
    os.environ.get('POLYGON_API_SECRET', '').strip(),
)


def _get_source_type(code: CodeItem):
    return None


def _get_solution_tag(solution: Solution, is_first: bool = False) -> api.SolutionTag:
    if solution.outcome == ExpectedOutcome.ACCEPTED:
        return api.SolutionTag.OK if not is_first else api.SolutionTag.MA
    if solution.outcome == ExpectedOutcome.ACCEPTED_OR_TLE:
        return api.SolutionTag.TO
    if solution.outcome == ExpectedOutcome.WRONG_ANSWER:
        return api.SolutionTag.WA
    if solution.outcome == ExpectedOutcome.TIME_LIMIT_EXCEEDED:
        return api.SolutionTag.TL
    if solution.outcome == ExpectedOutcome.MEMORY_LIMIT_EXCEEDED:
        return api.SolutionTag.ML
    if solution.outcome == ExpectedOutcome.RUNTIME_ERROR:
        return api.SolutionTag.RE
    return api.SolutionTag.RJ


def _find_or_create_problem(problem_name: str) -> api.Problem:
    results = POLY.problems_list(name=problem_name)
    for result in results:
        if result.name == problem_name:
            console.console.print(
                f'Found already existing problem [item]{problem_name}[/item].'
            )
            return result
    console.console.print(f'Creating new problem [item]{problem_name}[/item].')
    return POLY.problem_create(problem_name)


def _update_problem_info(problem: api.Problem):
    pkg = package.find_problem_package_or_die()

    problem.update_info(
        api.ProblemInfo(
            interactive=pkg.type == TaskType.COMMUNICATION,
            time_limit=pkg.timeLimit,
            memory_limit=pkg.memoryLimit,
        )
    )


def _get_checker_name() -> str:
    checker = package.get_checker()
    return checker.path.with_stem('checker').name


def _get_interactor_name() -> str:
    interactor = package.get_interactor()
    return interactor.path.with_stem('interactor').name


def _get_validator_name() -> str:
    validator = package.get_validator()
    return validator.path.with_stem('validator').name


def _update_rbx_header(problem: api.Problem):
    console.console.print('Uploading rbx.h...')
    rbx_header = header.get_header()
    problem.save_file(
        type=api.FileType.RESOURCE,
        name='rbx.h',
        file=rbx_header.read_bytes(),
        source_type=None,
    )


def _update_checker(problem: api.Problem):
    console.console.print('Uploading checker...')
    checker = package.get_checker()
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_checker_name(),
        file=checker.path.read_bytes(),
        source_type=_get_source_type(checker),
    )

    problem.set_checker(_get_checker_name())


def _update_interactor(problem: api.Problem):
    console.console.print('Uploading interactor...')
    interactor = package.get_interactor()
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_interactor_name(),
        file=interactor.path.read_bytes(),
        source_type=_get_source_type(interactor),
    )

    problem.set_interactor(_get_interactor_name())


def _upload_validator(problem: api.Problem):
    console.console.print('Uploading validator...')
    validator = package.get_validator()
    problem.save_file(
        type=api.FileType.SOURCE,
        name=_get_validator_name(),
        file=validator.path.read_bytes(),
        source_type=_get_source_type(validator),
    )

    problem.set_validator(_get_validator_name())


def _save_skip_coinciding_testcases(problem: api.Problem, *args, **kwargs) -> bool:
    try:
        problem.save_test(*args, **kwargs)
    except api.PolygonRequestFailedException as e:
        if 'test coincides with' in e.comment.lower():
            return False
        raise
    return True


def _get_test_params_for_statement(
    testcase: Testcase, is_sample: bool
) -> Dict[str, Any]:
    if not is_sample:
        return {}
    res: Dict[str, Any] = {'test_use_in_statements': True}
    if testcase.outputPath is not None:
        res['test_output_for_statements'] = testcase.outputPath.read_text()
    else:
        return res

    pio_path = testcase.outputPath.with_suffix('.pio')
    if pio_path.is_file():
        interaction = parse_interaction(pio_path)
        res['test_input_for_statements'], res['test_output_for_statements'] = (
            get_alternate_interaction_texts(interaction)
        )
    else:
        pin_path = testcase.outputPath.with_suffix('.pin')
        if pin_path.is_file():
            res['test_input_for_statements'] = pin_path.read_text()
        pout_path = testcase.outputPath.with_suffix('.pout')
        if pout_path.is_file():
            res['test_output_for_statements'] = pout_path.read_text()
    return res


def _upload_testcases(problem: api.Problem):
    pkg = package.find_problem_package_or_die()
    testcases = get_all_built_testcases()
    i = 0

    with rich.progress.Progress(speed_estimate_period=5) as progress:
        total_len = 0
        for group in pkg.testcases:
            total_len += len(testcases[group.name])
        task_id = progress.add_task('Uploading testcases...', total=total_len)
        for group in pkg.testcases:
            for testcase in testcases[group.name]:
                is_sample = group.name == 'samples'
                saved = _save_skip_coinciding_testcases(
                    problem,
                    testset='tests',
                    test_index=i + 1,
                    test_input=testcase.inputPath.read_text(),
                    **_get_test_params_for_statement(testcase, is_sample),
                )
                progress.update(task_id, advance=1)
                if saved:
                    i += 1


def _upload_solutions(problem: api.Problem):
    console.console.print('Uploading main solution...')
    pkg = package.find_problem_package_or_die()
    main_solution = pkg.solutions[0]
    if main_solution is None or main_solution.outcome != ExpectedOutcome.ACCEPTED:
        return
    problem.save_solution(
        main_solution.path.name,
        main_solution.path.read_bytes(),
        source_type=_get_source_type(main_solution),
        tag=api.SolutionTag.MA,
    )

    for i, solution in enumerate(pkg.solutions):
        console.console.print(
            f'Uploading solution [item]{solution.path.name}[/item] (tag: [item]{_get_solution_tag(solution, is_first=i == 0)}[/item])...'
        )
        problem.save_solution(
            solution.path.name,
            solution.path.read_bytes(),
            source_type=_get_source_type(solution),
            tag=_get_solution_tag(solution, is_first=i == 0),
        )


def _get_statement_for_language(language: str) -> Optional[Statement]:
    pkg = package.find_problem_package_or_die()
    for statement in pkg.statements:
        if statement.language == language:
            return statement
    return None


def _get_statement_blocks(statement: Statement) -> StatementBlocks:
    # TODO: actually try to convert to rbxTeX
    assert statement.type == StatementType.rbxTeX
    builder_problem = StatementBuilderProblem(
        package=package.find_problem_package_or_die(),
        statement=statement,
    )
    with tempfile.TemporaryDirectory() as temp_dir:
        return render_jinja_blocks(
            pathlib.Path(temp_dir),
            statement.path.read_bytes(),
            **builder_problem.build_inner_jinja_kwargs(),
        )


def _get_explanations(explanations: Dict[int, str]) -> str:
    entries = []
    for i in sorted(explanations):
        explanation = explanations[i]
        entries.append(f'\\textbf{{Explanation for example {i + 1}}}\n\n{explanation}')
    return '\n\n'.join(entries)


def _get_notes_with_explanations(blocks: StatementBlocks) -> Optional[str]:
    notes = blocks.blocks.get('notes')
    explanations = blocks.explanations
    if notes is None and not explanations:
        return None
    if notes is None:
        return _get_explanations(blocks.explanations)
    return notes + '\n\n' + _get_explanations(blocks.explanations)


def _upload_statement_resources(problem: api.Problem, statement: Statement):
    assets = get_relative_assets(statement.path, statement.assets)
    for asset, relative_asset in assets:
        console.console.print(
            f'Uploading statement resource [item]{relative_asset}[/item]...'
        )
        resource_bytes = asset.read_bytes()
        if len(resource_bytes) >= 1024 * 1024:  # >= 1mb
            console.console.print(
                f'[error]Statement resource [item]{relative_asset}[/item] is too large to upload (more than 1MB).[/error]'
            )
            raise typer.Exit(1)
        problem.save_statement_resource(
            name=str(relative_asset),
            file=resource_bytes,
        )


def _upload_statement(problem: api.Problem):
    pkg = package.find_problem_package_or_die()

    languages = set()
    for statement in pkg.statements:
        if not is_valid_lang_code(statement.language):
            continue
        languages.add(statement.language)
    for language in languages:
        statement = _get_statement_for_language(language)
        if statement is None:
            continue
        if statement.type != StatementType.rbxTeX:
            continue
        console.console.print(
            f'Uploading statement for language [item]{language}[/item] (polygon language: [item]{code_to_langs([language])[0]}[/item])...'
        )
        blocks = _get_statement_blocks(statement)
        polygon_statement = api.Statement(
            encoding='utf-8',
            name=statement.title,
            legend=blocks.blocks.get('legend') or '',
            input=blocks.blocks.get('input') or '',
            output=blocks.blocks.get('output') or '',
            interaction=(blocks.blocks.get('interaction') or '')
            if pkg.type == TaskType.COMMUNICATION
            else None,
            notes=_get_notes_with_explanations(blocks) or '',
        )
        problem.save_statement(
            lang=code_to_langs([language])[0], problem_statement=polygon_statement
        )

        _upload_statement_resources(problem, statement)


def _normalize_problem_name(name: str) -> str:
    return name.replace(' ', '-').replace('_', '-').lower()


async def upload_problem(name: str):
    pkg = package.find_problem_package_or_die()
    name = _normalize_problem_name(name)
    problem = _find_or_create_problem(name)
    _update_problem_info(problem)
    _update_rbx_header(problem)
    _update_checker(problem)

    if (
        pkg.type == TaskType.COMMUNICATION
        and package.get_interactor_or_nil() is not None
    ):
        _update_interactor(problem)

    # if pkg.validator is not None:
    #     _upload_validator(problem)

    _upload_solutions(problem)
    _upload_testcases(problem)
    _upload_statement(problem)

    # Commit.
    console.console.print('Committing changes...')
    problem.commit_changes()

    console.console.print(
        f'[success]Problem [item]{name}[/item] uploaded successfully![/success]'
    )



from typing import List, Optional

from pydantic import BaseModel

from rbx.box.presets.schema import TrackedAsset


class LockedAsset(TrackedAsset):
    hash: str


class PresetLock(BaseModel):
    name: str
    uri: Optional[str] = None

    @property
    def preset_name(self) -> str:
        return self.name

    assets: List[LockedAsset] = []



import pathlib
import re
from typing import Optional

from pydantic import BaseModel


class PresetFetchInfo(BaseModel):
    # The actual name of this preset.
    name: str

    # The URI to associate with this preset.
    uri: Optional[str] = None

    # The actual URI from where to fetch the repo.
    fetch_uri: Optional[str] = None

    # Inner directory from where to pull the preset.
    inner_dir: str = ''

    def is_remote(self) -> bool:
        return self.fetch_uri is not None

    def is_local_dir(self) -> bool:
        return bool(self.inner_dir) and not self.is_remote()


def get_preset_fetch_info(uri: Optional[str]) -> Optional[PresetFetchInfo]:
    if uri is None:
        return None

    def get_github_fetch_info(s: str) -> Optional[PresetFetchInfo]:
        pattern = r'(https:\/\/(?:[\w\-]+\.)?github\.com\/([\w\-]+\/[\w\.\-]+))(?:\.git)?(?:\/(.*))?'
        compiled = re.compile(pattern)
        match = compiled.match(s)
        if match is None:
            return None
        return PresetFetchInfo(
            name=match.group(2),
            uri=match.group(0),
            fetch_uri=match.group(1),
            inner_dir=match.group(3) or '',
        )

    def get_short_github_fetch_info(s: str) -> Optional[PresetFetchInfo]:
        pattern = r'(?:\@gh/)?([\w\-]+\/[\w\.\-]+)(?:\/(.*))?'
        compiled = re.compile(pattern)
        match = compiled.match(s)
        if match is None:
            return None
        return PresetFetchInfo(
            name=match.group(1),
            uri=match.group(0),
            fetch_uri=f'https://github.com/{match.group(1)}',
            inner_dir=match.group(2) or '',
        )

    def get_local_dir_fetch_info(s: str) -> Optional[PresetFetchInfo]:
        try:
            path = pathlib.Path(s)
            if not path.exists():
                return None
        except Exception:
            return None
        return PresetFetchInfo(name=path.name, inner_dir=str(path))

    def get_local_fetch_info(s: str) -> Optional[PresetFetchInfo]:
        pattern = r'[\w\-]+'
        compiled = re.compile(pattern)
        match = compiled.match(s)
        if match is None:
            return None
        return PresetFetchInfo(name=s)

    extractors = [
        get_github_fetch_info,
        get_short_github_fetch_info,
        get_local_dir_fetch_info,
        get_local_fetch_info,
    ]

    for extract in extractors:
        res = extract(uri)
        if res is not None:
            return res

    return None



from typing import List, Optional

from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import Screen
from textual.widgets import Footer, Header, Label, ListItem, ListView, RichLog

from rbx.box.testcase_extractors import (
    GenerationTestcaseEntry,
    extract_generation_testcases_from_groups,
)
from rbx.box.ui.widgets.file_log import FileLog
from rbx.box.ui.widgets.rich_log_box import RichLogBox
from rbx.box.ui.widgets.test_output_box import TestBoxWidget, TestcaseRenderingData


class TestExplorerScreen(Screen):
    BINDINGS = [
        ('q', 'app.pop_screen', 'Quit'),
        ('m', 'toggle_metadata', 'Toggle metadata'),
        ('1', 'show_output', 'Show output'),
        ('2', 'show_stderr', 'Show stderr'),
        ('3', 'show_log', 'Show log'),
    ]

    def __init__(self):
        super().__init__()
        self._entries: List[GenerationTestcaseEntry] = []

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        with Horizontal(id='test-explorer'):
            with Vertical(id='test-list-container'):
                yield ListView(id='test-list')
            with Vertical(id='test-details'):
                yield FileLog(id='test-input')
                yield TestBoxWidget(id='test-output')
                yield RichLogBox(id='test-metadata')

    async def on_mount(self):
        self.query_one('#test-list').border_title = 'Tests'
        self.query_one('#test-input').border_title = 'Input'

        metadata = self.query_one('#test-metadata', RichLogBox)
        metadata.display = False
        metadata.border_title = 'Metadata'
        metadata.wrap = True
        metadata.markup = True
        metadata.clear().write('No test selected')
        await self._update_tests()

    def action_toggle_metadata(self):
        metadata = self.query_one('#test-metadata', RichLogBox)
        metadata.display = not metadata.display

    def _update_selected_test(self, index: Optional[int]):
        input = self.query_one('#test-input', FileLog)
        output = self.query_one('#test-output', TestBoxWidget)
        metadata = self.query_one('#test-metadata', RichLog)

        if index is None:
            input.path = None
            output.reset()
            metadata.clear().write('No test selected')
            return
        entry = self._entries[index]
        input.path = entry.metadata.copied_to.inputPath

        output.data = TestcaseRenderingData.from_one_path(
            entry.metadata.copied_to.outputPath
        )

        metadata.clear()
        metadata.write(
            f'[bold]{entry.group_entry.group}[/bold] / [bold]{entry.group_entry.index}[/bold]'
        )
        if entry.metadata.copied_from is not None:
            metadata.write(
                f'[bold]Copied from:[/bold] {entry.metadata.copied_from.inputPath}'
            )
        if entry.metadata.generator_call is not None:
            metadata.write(f'[bold]Gen. call:[/bold] {entry.metadata.generator_call}')
        if entry.metadata.generator_script is not None:
            metadata.write(
                f'[bold]Gen. script:[/bold] {entry.metadata.generator_script}'
            )

    async def _update_tests(self):
        self.watch(
            self.query_one('#test-list', ListView),
            'index',
            self._update_selected_test,
        )

        self._entries = await extract_generation_testcases_from_groups()

        test_names = [
            f'{entry.group_entry.group}/{entry.group_entry.index}'
            for entry in self._entries
        ]

        await self.query_one('#test-list', ListView).clear()
        await self.query_one('#test-list', ListView).extend(
            [ListItem(Label(name)) for name in test_names]
        )

    def action_show_output(self):
        self.query_one('#test-output', TestBoxWidget).show_output()

    def action_show_stderr(self):
        self.query_one('#test-output', TestBoxWidget).show_stderr()

    def action_show_log(self):
        self.query_one('#test-output', TestBoxWidget).show_log()



import importlib.resources
import pathlib

import rich.markup
import rich.text
import rich.tree
from rich.filesize import decimal

from rbx import console


def get_testdata_path() -> pathlib.Path:
    with importlib.resources.as_file(
        importlib.resources.files('rbx') / 'testdata' / 'compatible'
    ) as file:
        return file.parent


def clear_all_functools_cache():
    from rbx.box import environment, header, package

    pkgs = [environment, package, header]

    for pkg in pkgs:
        for fn in pkg.__dict__.values():
            if hasattr(fn, 'cache_clear'):
                fn.cache_clear()


def walk_directory(
    directory: pathlib.Path, tree: rich.tree.Tree, show_hidden: bool = False
) -> None:
    """Recursively build a Tree with directory contents."""
    # Sort dirs first then by filename
    paths = sorted(
        pathlib.Path(directory).iterdir(),
        key=lambda path: (path.is_file(), path.name.lower()),
    )
    for path in paths:
        # Remove hidden files
        if path.name.startswith('.') and not show_hidden:
            continue
        if path.is_dir():
            style = 'dim' if path.name.startswith('__') else ''
            branch = tree.add(
                f'[bold magenta]:open_file_folder: [link file://{path}]{rich.markup.escape(path.name)}',
                style=style,
                guide_style=style,
            )
            walk_directory(path, branch)
        else:
            text_filename = rich.text.Text(path.name, 'green')
            text_filename.highlight_regex(r'\..*$', 'bold red')
            text_filename.stylize(f'link file://{path}')
            file_size = path.stat().st_size
            text_filename.append(f' ({decimal(file_size)})', 'blue')
            icon = '🐍 ' if path.suffix == '.py' else '📄 '
            tree.add(rich.text.Text(icon) + text_filename)


def print_directory_tree(directory: pathlib.Path, show_hidden: bool = False):
    tree = rich.tree.Tree(directory.name)
    walk_directory(directory, tree, show_hidden=show_hidden)
    console.console.print(tree)



import pathlib
import shutil
from typing import Annotated, Optional

import typer

from rbx import console, utils
from rbx.box import package, presets
from rbx.box.presets.fetch import get_preset_fetch_info


def create(
    name: Annotated[
        str,
        typer.Argument(
            help='The name of the problem package to create. This will also be the name of the folder.'
        ),
    ],
    preset: Annotated[
        Optional[str],
        typer.Option(
            '--preset',
            '-p',
            help='Which preset to use to create this package. Can be a named of an already installed preset, or an URI, in which case the preset will be downloaded.',
        ),
    ] = None,
    path: Optional[pathlib.Path] = None,
):
    preset = preset or 'default'
    console.console.print(f'Creating new problem [item]{name}[/item]...')

    fetch_info = get_preset_fetch_info(preset)
    if fetch_info is None:
        console.console.print(
            f'[error]Invalid preset name/URI [item]{preset}[/item].[/error]'
        )
        raise typer.Exit(1)

    if fetch_info.fetch_uri is not None:
        preset = presets.install_from_remote(fetch_info)

    preset_cfg = presets.get_installed_preset(preset)

    problem_path = (
        presets.get_preset_installation_path(preset) / preset_cfg.problem
        if preset_cfg.problem is not None
        else presets.get_preset_installation_path('default') / 'problem'
    )

    if not problem_path.is_dir():
        console.console.print(
            f'[error]Problem template [item]{problem_path}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)

    dest_path = path or pathlib.Path(name)

    if dest_path.exists():
        console.console.print(
            f'[error]Directory [item]{dest_path}[/item] already exists.[/error]'
        )
        raise typer.Exit(1)

    dest_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copytree(str(problem_path), str(dest_path))

    # Remove a few left overs.
    shutil.rmtree(str(dest_path / 'build'), ignore_errors=True)
    shutil.rmtree(str(dest_path / '.box'), ignore_errors=True)
    for lock in dest_path.rglob('.preset-lock.yml'):
        lock.unlink(missing_ok=True)

    # Change problem name.
    ru, problem = package.get_ruyaml(dest_path)
    problem['name'] = name
    utils.save_ruyaml(dest_path / 'problem.rbx.yml', ru, problem)

    presets.generate_lock(preset, root=dest_path)



import typing
from typing import List, Optional

from rbx import utils
from rbx.box import package, solutions
from rbx.box.solutions import SolutionReportSkeleton, SolutionSkeleton
from rbx.box.testcase_utils import TestcaseEntry
from rbx.grading.steps import Evaluation


def has_run() -> bool:
    return (package.get_problem_runs_dir() / 'skeleton.yml').is_file()


def get_skeleton() -> SolutionReportSkeleton:
    skeleton_path = package.get_problem_runs_dir() / 'skeleton.yml'
    return utils.model_from_yaml(
        SolutionReportSkeleton,
        skeleton_path.read_text(),
    )


def get_solution_eval(
    solution: SolutionSkeleton, entry: TestcaseEntry
) -> Optional[Evaluation]:
    path = solution.get_entry_prefix(entry).with_suffix('.eval')
    if not path.is_file():
        return None
    return utils.model_from_yaml(Evaluation, path.read_text())


def get_solution_evals(
    skeleton: SolutionReportSkeleton, solution: SolutionSkeleton
) -> List[Optional[Evaluation]]:
    return [get_solution_eval(solution, entry) for entry in skeleton.entries]


def get_solution_evals_or_null(
    skeleton: SolutionReportSkeleton, solution: SolutionSkeleton
) -> Optional[List[Evaluation]]:
    evals = get_solution_evals(skeleton, solution)
    if any(e is None for e in evals):
        return None
    return typing.cast(List[Evaluation], evals)


def get_solution_markup(
    skeleton: SolutionReportSkeleton, solution: SolutionSkeleton
) -> str:
    header = f'[b $accent]{solution.path}[/b $accent] ({solution.path})'

    evals = get_solution_evals_or_null(skeleton, solution)
    report = solutions.get_solution_outcome_report(
        solution, evals or [], skeleton.verification, subset=False
    )
    if evals is None:
        return header + '\n' + report.get_verdict_markup(incomplete=True)
    return header + '\n' + report.get_outcome_markup()


def get_run_testcase_markup(solution: SolutionSkeleton, entry: TestcaseEntry) -> str:
    eval = get_solution_eval(solution, entry)
    if eval is None:
        return f'{entry}'
    testcase_markup = solutions.get_testcase_markup_verdict(eval)
    return f'{testcase_markup} {entry}'


def _get_checker_msg_last_line(eval: Evaluation) -> Optional[str]:
    if eval.result.message is None:
        return None
    return eval.result.message.rstrip().split('\n')[-1]


def get_run_testcase_metadata_markup(
    skeleton: SolutionReportSkeleton, solution: SolutionSkeleton, entry: TestcaseEntry
) -> Optional[str]:
    eval = get_solution_eval(solution, entry)
    if eval is None:
        return None
    time_str = solutions.get_capped_evals_formatted_time(
        solution, [eval], skeleton.verification
    )
    memory_str = solutions.get_evals_formatted_memory([eval])

    checker_msg = _get_checker_msg_last_line(eval)

    lines = []
    lines.append(
        f'[b]{solutions.get_full_outcome_markup_verdict(eval.result.outcome)}[/b]'
    )
    lines.append(f'[b]Time:[/b] {time_str} / [b]Memory:[/b] {memory_str}')
    if checker_msg is not None:
        lines.append(f'[b]Checker:[/b] {checker_msg}')
    return '\n'.join(lines)



def get_formatted_memory(memory_in_bytes: int, mib_decimal_places: int = 0) -> str:
    if memory_in_bytes < 1024 * 1024:
        if memory_in_bytes < 1024:
            return f'{memory_in_bytes} B'
        return f'{memory_in_bytes / 1024:.0f} KiB'
    return f'{memory_in_bytes / (1024 * 1024):.{mib_decimal_places}f} MiB'


def get_formatted_time(time_in_ms: int) -> str:
    return f'{time_in_ms} ms'



from typing import Any, Dict, List, Optional, Tuple

from rbx.grading import steps
from rbx.grading.caching import DependencyCache, NoCacheException
from rbx.grading.judge.sandbox import SandboxBase, SandboxParams
from rbx.grading.steps import (
    GradingArtifacts,
    GradingLogsHolder,
    RunLog,
    RunLogMetadata,
)


def _get_prefixed_cacheable_params(
    params: Dict[str, Any], prefix: str
) -> Dict[str, Any]:
    return {f'{prefix}.{k}': v for k, v in params.items()}


def compile(
    commands: List[str],
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
    dependency_cache: DependencyCache,
):
    artifacts.logs = GradingLogsHolder()

    ok = True
    with dependency_cache(
        commands, [artifacts], params.get_cacheable_params()
    ) as is_cached:
        if not is_cached and not steps.compile(
            commands=commands,
            params=params,
            artifacts=artifacts,
            sandbox=sandbox,
        ):
            ok = False
            raise NoCacheException()

    return ok


async def run(
    command: str,
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
    dependency_cache: DependencyCache,
    metadata: Optional[RunLogMetadata] = None,
) -> Optional[RunLog]:
    artifacts.logs = GradingLogsHolder()

    cacheable_params = params.get_cacheable_params()
    if metadata is not None and metadata.retryIndex is not None:
        cacheable_params['__retry_index__'] = metadata.retryIndex

    with dependency_cache([command], [artifacts], cacheable_params) as is_cached:
        if not is_cached:
            await steps.run(
                command=command,
                params=params,
                artifacts=artifacts,
                sandbox=sandbox,
                metadata=metadata,
            )

    return artifacts.logs.run


async def run_coordinated(
    interactor: steps.CoordinatedRunParams,
    solution: steps.CoordinatedRunParams,
    dependency_cache: DependencyCache,
) -> Tuple[Optional[RunLog], Optional[RunLog]]:
    interactor.artifacts.logs = GradingLogsHolder()
    solution.artifacts.logs = GradingLogsHolder()

    cacheable_params = {
        **_get_prefixed_cacheable_params(
            interactor.params.get_cacheable_params(), 'interactor'
        ),
        **_get_prefixed_cacheable_params(
            solution.params.get_cacheable_params(), 'solution'
        ),
    }

    if interactor.metadata is not None and interactor.metadata.retryIndex is not None:
        cacheable_params['interactor.__retry_index__'] = interactor.metadata.retryIndex
    if solution.metadata is not None and solution.metadata.retryIndex is not None:
        cacheable_params['solution.__retry_index__'] = solution.metadata.retryIndex

    with dependency_cache(
        [interactor.command, solution.command],
        [interactor.artifacts, solution.artifacts],
        cacheable_params,
    ) as is_cached:
        if not is_cached:
            await steps.run_coordinated(interactor, solution)

    return (
        interactor.artifacts.logs.run,
        solution.artifacts.logs.run,
    )



import pathlib
import shlex
from typing import Dict, Iterable, List, Optional, Set, Tuple

import typer
from pydantic import BaseModel

from rbx import console
from rbx.box import package
from rbx.box.code import SanitizationLevel, compile_item, run_item
from rbx.box.schema import CodeItem, Primitive
from rbx.box.testcase_extractors import (
    GenerationTestcaseEntry,
    extract_generation_testcases_from_groups,
)
from rbx.grading.judge.sandbox import SandboxBase
from rbx.grading.steps import (
    DigestHolder,
    DigestOrDest,
    DigestOrSource,
    GradingFileOutput,
)
from rbx.utils import StatusProgress

HitBounds = Dict[str, Tuple[bool, bool]]


class TestcaseValidationInfo(BaseModel):
    validator: CodeItem
    group: str
    path: pathlib.Path
    ok: bool
    hit_bounds: HitBounds
    message: Optional[str] = None


def _compile_validator(validator: CodeItem) -> str:
    try:
        digest = compile_item(validator, sanitized=SanitizationLevel.PREFER)
    except:
        console.console.print(
            f'[error]Failed compiling validator [item]{validator.path}[/item][/error]'
        )
        raise
    return digest


def _bounds_or(lhs: Tuple[bool, bool], rhs: Tuple[bool, bool]) -> Tuple[bool, bool]:
    return (lhs[0] or rhs[0], lhs[1] or rhs[1])


def _process_bounds(log: str) -> HitBounds:
    bounds: HitBounds = {}
    for line in log.splitlines():
        items = line.split(':')
        if len(items) != 2:
            continue
        k, v = items
        v = v.strip()

        if 'constant-bounds' in k:
            continue

        hit = ('min-value-hit' in v, 'max-value-hit' in v)
        if k not in bounds:
            bounds[k] = hit
            continue
        bounds[k] = _bounds_or(bounds[k], hit)
    return bounds


def _merge_hit_bounds(hit_bounds: Iterable[HitBounds]) -> HitBounds:
    res: HitBounds = {}
    for hb in hit_bounds:
        for k, hit in hb.items():
            if k not in res:
                res[k] = hit
                continue
            res[k] = _bounds_or(res[k], hit)
    return res


def _has_group_specific_validator() -> bool:
    pkg = package.find_problem_package_or_die()

    return any(group.validator is not None for group in pkg.testcases)


async def _validate_testcase(
    testcase: pathlib.Path,
    validator: CodeItem,
    validator_digest: str,
    vars: Optional[Dict[str, Primitive]] = None,
) -> Tuple[bool, Optional[str], HitBounds]:
    vars = vars or {}
    for var in vars:
        assert (
            var.isidentifier()
        ), f'Variable {var} should be a valid Python identifier.'
    # TODO: check if needs to do some escaping
    var_args = [f'--{k}={v}' for k, v in vars.items()]
    var_args.extend(['--testOverviewLogFileName', 'validator.log'])

    message_digest = DigestHolder()
    log_digest = DigestHolder()
    run_log = await run_item(
        validator,
        DigestOrSource.create(validator_digest),
        stdin=DigestOrSource.create(testcase),
        stderr=DigestOrDest.create(message_digest),
        outputs=[
            GradingFileOutput(
                src=pathlib.Path('validator.log'),
                digest=log_digest,
                optional=True,
            )
        ],
        extra_args=shlex.join(var_args) if var_args else None,
    )

    if (
        run_log is not None
        and run_log.exitcode != 0
        and run_log.exitstatus != SandboxBase.EXIT_NONZERO_RETURN
    ):
        console.console.print(
            f'[error]Validator [item]{validator.path}[/item] failed unexpectedly.[/error]'
        )
        console.console.print(f'[error]Summary:[/error] {run_log.get_summary()}')
        raise typer.Exit(1)

    log_overview = ''
    if log_digest.value is not None:
        log_overview = package.get_digest_as_string(log_digest.value or '')
    message = package.get_digest_as_string(message_digest.value or '')
    return (
        run_log is not None and run_log.exitcode == 0,
        message,
        _process_bounds(log_overview or ''),
    )


async def _validate_test(
    testcase: pathlib.Path,
    validator: CodeItem,
    validator_digest: str,
) -> Tuple[bool, Optional[str], HitBounds]:
    pkg = package.find_problem_package_or_die()
    return await _validate_testcase(
        testcase, validator, validator_digest, vars=pkg.expanded_vars
    )


def compile_main_validator() -> Optional[Tuple[CodeItem, str]]:
    pkg = package.find_problem_package_or_die()
    if pkg.validator is None:
        return None

    return pkg.validator, _compile_validator(pkg.validator)


async def validate_one_off(
    testcase: pathlib.Path,
    validator: CodeItem,
    validator_digest: str,
) -> TestcaseValidationInfo:
    ok, message, _ = await _validate_test(testcase, validator, validator_digest)
    info = TestcaseValidationInfo(
        validator=validator,
        group='interactive',
        path=testcase,
        ok=ok,
        hit_bounds={},
        message=message,
    )
    return info


def compile_validators(
    validators: List[CodeItem],
    progress: Optional[StatusProgress] = None,
) -> Dict[str, str]:
    validator_to_compiled_digest = {}

    validator_to_compiled_digest = {}

    for validator in validators:
        if str(validator.path) in validator_to_compiled_digest:
            continue

        if progress:
            progress.update(f'Compiling validator [item]{validator.path}[/item]...')
        validator_to_compiled_digest[str(validator.path)] = _compile_validator(
            validator
        )

    return validator_to_compiled_digest


def compile_validators_for_entries(
    validation_entries: List[GenerationTestcaseEntry],
    progress: Optional[StatusProgress] = None,
) -> Dict[str, str]:
    validators = []

    for entry in validation_entries:
        if entry.validator is not None:
            validators.append(entry.validator)
        validators.extend(entry.extra_validators)

    return compile_validators(validators, progress=progress)


async def validate_testcases(
    progress: Optional[StatusProgress] = None,
    groups: Optional[Set[str]] = None,
) -> List[TestcaseValidationInfo]:
    def step():
        if progress is not None:
            progress.step()

    validation_entries = await extract_generation_testcases_from_groups(groups)
    validator_to_compiled_digest = compile_validators_for_entries(
        validation_entries, progress=progress
    )

    validation_info = []

    for entry in validation_entries:
        input_path = entry.metadata.copied_to.inputPath
        if not input_path.is_file():
            continue

        # Main validation.
        if entry.validator is not None:
            compiled_digest = validator_to_compiled_digest[str(entry.validator.path)]
            ok, message, hit_bounds = await _validate_test(
                input_path, entry.validator, compiled_digest
            )
            validation_info.append(
                TestcaseValidationInfo(
                    validator=entry.validator,
                    group=entry.group_entry.group,
                    path=input_path,
                    ok=ok,
                    hit_bounds=hit_bounds,
                    message=message,
                )
            )

        for extra_validator in entry.extra_validators:
            compiled_digest = validator_to_compiled_digest[str(extra_validator.path)]
            ok, message, hit_bounds = await _validate_test(
                input_path, extra_validator, compiled_digest
            )
            validation_info.append(
                TestcaseValidationInfo(
                    validator=extra_validator,
                    group=entry.group_entry.group,
                    path=input_path,
                    ok=ok,
                    hit_bounds=hit_bounds,
                    message=message,
                )
            )

        step()

    return validation_info


def has_validation_errors(infos: List[TestcaseValidationInfo]) -> bool:
    return any(not info.ok for info in infos)


def print_validation_report(infos: List[TestcaseValidationInfo]):
    console.console.rule('Validation report', style='status')
    hit_bounds_per_group: Dict[Optional[str], HitBounds] = {}
    any_failure = False
    for info in infos:
        if not info.ok:
            console.console.print(
                f'[error]Testcase [item]{info.path}[/item] failed verification on validator [item]{info.validator.path}[/item]:[/error]'
            )
            console.console.print(info.message)
            any_failure = True
            continue

        if info.group not in hit_bounds_per_group:
            hit_bounds_per_group[info.group] = {}
        hit_bounds_per_group[info.group] = _merge_hit_bounds(
            [hit_bounds_per_group[info.group], info.hit_bounds]
        )

    if not hit_bounds_per_group:
        console.console.print()
        return

    if not _has_group_specific_validator():
        hit_bounds_per_group = {None: _merge_hit_bounds(hit_bounds_per_group.values())}

    def _is_hit_bound_good(hit_bounds: HitBounds) -> bool:
        return any(not v[0] or not v[1] for v in hit_bounds.values())

    # Cleanup entries in hit bounds per group that are totally empty.
    # Also skip samples.
    hit_bounds_per_group = {
        k: v
        for k, v in hit_bounds_per_group.items()
        if _is_hit_bound_good(v) and k != 'samples'
    }

    all_groups = set(info.group for info in infos)
    if len(all_groups) == 1 and 'samples' in all_groups:
        # If there's only the samples group, do not check for hit bounds.
        hit_bounds_per_group = {}

    if not hit_bounds_per_group and not any_failure:
        console.console.print('[info]No validation issues found.[/info]')
        return

    for group, hit_bounds in hit_bounds_per_group.items():
        if group is None:
            console.console.print('Hit bounds:')
        else:
            console.console.print(f'Group [item]{group}[/item] hit bounds:')

        for k, v in hit_bounds.items():
            if all(v):
                continue

            if not v[0]:
                console.console.print(
                    f'  - {k}: [warning]min-value not hit[/warning]',
                )
            if not v[1]:
                console.console.print(
                    f'  - {k}: [warning]max-value not hit[/warning]',
                )
        console.console.print()



import pathlib

import pytest

from rbx.box.generators import generate_testcases
from rbx.box.validators import validate_testcases


@pytest.mark.test_pkg('box1')
async def test_validators(pkg_from_testdata: pathlib.Path):
    await generate_testcases()
    validation_infos = await validate_testcases()

    for info in validation_infos:
        assert info.ok



import contextlib
import functools
import pathlib
from typing import Optional

import typer

from rbx import console
from rbx.box.sanitizers import warning_stack
from rbx.utils import new_cd


def find_package(root: pathlib.Path = pathlib.Path()) -> Optional[pathlib.Path]:
    root = root.resolve()

    def has_file():
        problem_yaml_path = root / 'problem.rbx.yml'
        contest_yaml_path = root / 'contest.rbx.yml'
        return problem_yaml_path.is_file() or contest_yaml_path.is_file()

    while root != pathlib.PosixPath('/') and not has_file():
        root = root.parent
    if not has_file():
        return None
    return root


def is_problem_package(root: pathlib.Path = pathlib.Path()) -> bool:
    dir = find_package(root)
    if dir is None:
        return False
    return (dir / 'problem.rbx.yml').is_file()


def is_contest_package(root: pathlib.Path = pathlib.Path()) -> bool:
    dir = find_package(root)
    if dir is None:
        return False
    return (dir / 'contest.rbx.yml').is_file()


def within_closest_package(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        package = find_package()
        if package is None:
            console.console.print('[error]No rbx package found.[/error]')
            raise typer.Exit(1)
        # Get deepest package.
        with new_package_cd(package):
            return func(*args, **kwargs)

    return wrapper


@contextlib.contextmanager
def new_package_cd(x: pathlib.Path):
    with new_cd(x):
        yield
        warning_stack.print_warning_stack_report()



import dataclasses
import pathlib
import re
import resource
import shlex
from enum import Enum
from pathlib import PosixPath
from typing import List, Optional

import rich
import rich.text
import typer

from rbx import console
from rbx.box import download, package, setter_config, state
from rbx.box.environment import (
    ExecutionConfig,
    FileMapping,
    get_compilation_config,
    get_execution_config,
    get_file_mapping,
    get_language,
    get_mapped_command,
    get_mapped_commands,
    get_sandbox_params_from_config,
    merge_execution_configs,
)
from rbx.box.formatting import get_formatted_memory
from rbx.box.sanitizers import warning_stack
from rbx.box.schema import CodeItem
from rbx.grading import steps, steps_with_caching
from rbx.grading.judge.sandbox import SandboxParams
from rbx.grading.steps import (
    DigestHolder,
    DigestOrDest,
    DigestOrSource,
    GradingArtifacts,
    GradingFileInput,
    GradingFileOutput,
    RunLog,
    RunLogMetadata,
    is_cxx_command,
)


class SanitizationLevel(Enum):
    NONE = 0
    PREFER = 1
    FORCE = 2

    def should_sanitize(self) -> bool:
        cfg = setter_config.get_setter_config()
        if cfg.sanitizers.enabled or state.STATE.sanitized:
            return self.value >= SanitizationLevel.PREFER.value
        return self.value >= SanitizationLevel.FORCE.value


def substitute_commands(commands: List[str], sanitized: bool = False) -> List[str]:
    cfg = setter_config.get_setter_config()
    return [cfg.substitute_command(command, sanitized) for command in commands]


def get_extension(code: CodeItem) -> str:
    path: pathlib.Path = PosixPath(code.path)
    return path.suffix[1:]


def find_language_name(code: CodeItem) -> str:
    if code.language is not None:
        return get_language(code.language).name
    return get_language(get_extension(code)).name


def is_executable_sanitized(executable: DigestOrSource) -> bool:
    if executable.digest is None:
        return False
    storage = package.get_cache_storage()
    return storage.exists(f'{executable.digest.value}.san')


def add_sanitizer_flags_to_command(command: str) -> str:
    if is_cxx_command(command):
        return command + ' -fsanitize=address,undefined -fno-omit-frame-pointer -g'
    return command


def add_sanitizer_flags(commands: List[str]) -> List[str]:
    return [add_sanitizer_flags_to_command(command) for command in commands]


CXX_WARNING_FLAGS = (
    '-Wall -Wshadow -Wno-unused-result -Wno-sign-compare -Wno-char-subscripts'
)


def add_warning_flags_to_command(command: str) -> str:
    if is_cxx_command(command):
        return command + ' ' + CXX_WARNING_FLAGS
    return command


def add_warning_flags(commands: List[str], force_warnings: bool) -> List[str]:
    cfg = setter_config.get_setter_config()
    if cfg.warnings.enabled or force_warnings:
        return [add_warning_flags_to_command(command) for command in commands]
    return commands


def _add_warning_pragmas(code: str) -> str:
    flags = CXX_WARNING_FLAGS.split()
    pragma_lines = '\n'.join(
        [
            f'#pragma GCC diagnostic ignored "{flag}"'
            for flag in flags
            if not flag.startswith('-Wno-')
        ]
    )

    return re.sub(
        r'^(#include[^\n]*)',
        '#pragma GCC diagnostic push\n'
        + pragma_lines
        + '\n\\1'
        + '\n#pragma GCC diagnostic pop\n',
        code,
        flags=re.MULTILINE,
    )


def _add_warning_pragmas_around(code: str) -> str:
    flags = CXX_WARNING_FLAGS.split()
    pragma_lines = '\n'.join(
        [
            f'#pragma GCC diagnostic ignored "{flag}"'
            for flag in flags
            if not flag.startswith('-Wno-')
        ]
    )

    return (
        '#pragma GCC diagnostic push\n'
        + pragma_lines
        + '\n'
        + code
        + '\n'
        + '#pragma GCC diagnostic pop\n'
    )


def _ignore_warning_in_cxx_input(input: GradingFileInput):
    if input.src is None or input.src.suffix not in ('.h', '.hpp'):
        return
    preprocessed_path = package.get_problem_preprocessed_path(input.src)
    preprocessed_path.write_text(_add_warning_pragmas_around(input.src.read_text()))
    input.src = preprocessed_path


def _maybe_rename_java_class(
    compilable_path: pathlib.Path, file_mapping: FileMapping
) -> pathlib.Path:
    mapped_path = PosixPath(file_mapping.compilable)
    if mapped_path.suffix != '.java':
        return compilable_path
    import re

    cls_name = mapped_path.stem

    java_content = compilable_path.read_text()
    regex = re.compile(r'public\s+class\s+[A-Za-z0-9_$]+([^A-Za-z0-9_$])')
    match = regex.search(java_content)
    if match is None:
        console.console.print(
            f'[error]Java public class not found in file: [item]{compilable_path}[/item][/error]'
        )
        raise typer.Exit(1)

    new_content = regex.sub(f'public class {cls_name}\\1', java_content)
    if new_content == java_content:
        return compilable_path

    preprocessed_path = package.get_problem_preprocessed_path(compilable_path)
    preprocessed_path.write_text(new_content)
    return preprocessed_path


def _format_stack_limit(limit: int) -> str:
    if limit == resource.RLIM_INFINITY:
        return 'unlimited'
    return get_formatted_memory(limit)


def _check_stack_limit():
    if not state.STATE.run_through_cli:
        return
    soft, hard = resource.RLIM_INFINITY, resource.RLIM_INFINITY

    TARGET = 256 * 1024 * 1024  # 256 MiB
    try:
        soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
    except Exception:
        pass

    if soft != hard and soft != resource.RLIM_INFINITY and soft < TARGET:
        soft_fmt = _format_stack_limit(soft)
        hard_fmt = _format_stack_limit(hard)
        console.console.print(
            f'[error]Stack limit is too low (limit is set as [item]{soft_fmt}[/item], but configured user capacity is [item]{hard_fmt}[/item]).[/error]'
        )
        console.console.print(
            '[error]It is not safe to develop problems in [item]rbx[/item] with this configuration.[/error]'
        )
        console.console.print(
            'To solve this, add the following lines to the end of your [item]~/.bashrc[/item] or [item]~/.zshrc[/item] file (or equivalent shell configuration file):'
        )

        target_text = TARGET
        if hard != resource.RLIM_INFINITY:
            target_text = min(hard, TARGET)
        console.console.print(
            f"""
```
function rbx() {{
    local rbx_bin=`bash -c "type -P rbx"`
    ulimit -s {target_text // 1024} && $rbx_bin "$@"
}}
```
        """
        )
        console.console.print()
        console.console.print(
            'You can read more about this in [item]https://rsalesc.github.io/rbx/stack-limit/[/item].'
        )
        raise typer.Exit(1)


@dataclasses.dataclass
class PreparedRun:
    command: str
    sandbox_params: SandboxParams
    artifacts: GradingArtifacts
    sanitized: bool

    file_mapping: FileMapping
    metadata: RunLogMetadata


@dataclasses.dataclass
class CaptureSpec:
    prefix: str
    output: Optional[DigestOrDest] = None
    merged_capture: Optional[pathlib.Path] = None


def _prepare_for_communication(
    run: PreparedRun,
    stdin: pathlib.Path,
    stdout: pathlib.Path,
    reverse_io: bool = False,
    capture: Optional[CaptureSpec] = None,
):
    run.sandbox_params.set_stdio(
        stdin=stdin,
        stdout=stdout,
    )
    run.sandbox_params.reverse_io = reverse_io
    if capture is not None:
        run.sandbox_params.timeit_prefix = capture.prefix

        if capture.output is not None:
            output_path = PosixPath('capture')
            run.sandbox_params.timeit_dups['do'].append(output_path)

            run.artifacts.outputs.append(
                GradingFileOutput(
                    src=output_path,
                    **capture.output.expand(),
                    touch=True,
                )
            )

        if capture.merged_capture is not None:
            merged_output_path = package.get_merged_capture_path().resolve()
            run.sandbox_params.timeit_dups['Do'].append(merged_output_path)

            run.artifacts.outputs.append(
                GradingFileOutput(
                    src=merged_output_path,
                    dest=capture.merged_capture,
                )
            )


def _prepare_run(
    code: CodeItem,
    executable: DigestOrSource,
    stdin: Optional[DigestOrSource] = None,
    stdout: Optional[DigestOrDest] = None,
    stderr: Optional[DigestOrDest] = None,
    inputs: Optional[List[GradingFileInput]] = None,
    outputs: Optional[List[GradingFileOutput]] = None,
    extra_args: Optional[str] = None,
    extra_config: Optional[ExecutionConfig] = None,
    retry_index: Optional[int] = None,
):
    language = find_language_name(code)
    execution_options = get_execution_config(language)
    if extra_config is not None:
        execution_options = merge_execution_configs([execution_options, extra_config])
    file_mapping = get_file_mapping(language)
    sandbox_params = get_sandbox_params_from_config(execution_options.sandbox)

    # Sanitization parameters.
    sanitized = False
    if is_executable_sanitized(executable):
        # Remove any memory constraints for a sanitized executable.
        # Sanitizers are known to be memory-hungry.
        sandbox_params.address_space = None

        # Reset timeout configs since sanitizers are known to be time-hungry.
        sandbox_params.timeout = None
        sandbox_params.wallclock_timeout = None
        sanitized = True

    sandbox_params.set_stdall(
        stdin=PosixPath(file_mapping.input) if stdin is not None else None,
        stdout=PosixPath(file_mapping.output) if stdout is not None else None,
        stderr=PosixPath(file_mapping.error)
        if stderr is not None or sanitized
        else None,
    )

    assert execution_options.command
    command = get_mapped_command(execution_options.command, file_mapping)
    command = substitute_commands([command], sanitized=sanitized)[0]

    if extra_args is not None:
        splitted_command = shlex.split(command)
        splitted_command.extend(shlex.split(extra_args))
        command = shlex.join(splitted_command)

    artifacts = GradingArtifacts()
    artifacts.inputs.append(
        GradingFileInput(
            **executable.expand(),
            dest=PosixPath(file_mapping.executable),
            executable=True,
        )
    )
    if stdin is not None:
        artifacts.inputs.append(
            GradingFileInput(
                **stdin.expand(),
                dest=PosixPath(file_mapping.input),
            )
        )
    if stdout is not None:
        artifacts.outputs.append(
            GradingFileOutput(
                src=PosixPath(file_mapping.output),
                **stdout.expand(),
                touch=True,
            )
        )
    if stderr is not None:
        artifacts.outputs.append(
            GradingFileOutput(
                src=PosixPath(file_mapping.error),
                **stderr.expand(),
                touch=True,
            )
        )
    if inputs:
        artifacts.inputs.extend(inputs)
    if outputs:
        artifacts.outputs.extend(outputs)

    return PreparedRun(
        command=command,
        sandbox_params=sandbox_params,
        artifacts=artifacts,
        sanitized=sanitized,
        file_mapping=file_mapping,
        metadata=RunLogMetadata(
            language=code.language,
            is_sanitized=sanitized,
            timeLimit=sandbox_params.timeout,
            memoryLimit=sandbox_params.address_space,
            limits=execution_options.problemLimits,
            retryIndex=retry_index,
        ),
    )


# Compile code item and return its digest in the storage.
def compile_item(
    code: CodeItem,
    sanitized: SanitizationLevel = SanitizationLevel.PREFER,
    force_warnings: bool = False,
    verbose: bool = False,
) -> str:
    _check_stack_limit()

    compilable_path = PosixPath(code.path)

    if not compilable_path.is_file():
        console.console.print(
            f'[error]Compilation file not found: [item]{compilable_path}[/item][/error]'
        )
        raise typer.Exit(1)

    language = find_language_name(code)
    compilation_options = get_compilation_config(language)
    file_mapping = get_file_mapping(language)
    dependency_cache = package.get_dependency_cache()
    sandbox = package.get_singleton_sandbox()
    sandbox_params = get_sandbox_params_from_config(compilation_options.sandbox)

    if not compilation_options.commands:
        # Language is not compiled.
        return sandbox.file_cacher.put_file_from_path(compilable_path)

    commands = get_mapped_commands(compilation_options.commands, file_mapping)
    commands = add_warning_flags(commands, force_warnings)
    commands = substitute_commands(commands, sanitized=sanitized.should_sanitize())

    if sanitized.should_sanitize():
        commands = add_sanitizer_flags(commands)

        # Remove any memory constraints for a sanitized executable.
        # Sanitizers are known to be memory-hungry.
        sandbox_params.address_space = None

        # Reset timeout configs since sanitizers are known to be time-hungry.
        sandbox_params.timeout = None
        sandbox_params.wallclock_timeout = None

    compiled_digest = DigestHolder()

    artifacts = GradingArtifacts()
    artifacts.inputs.extend(
        GradingFileInput(src=src, dest=dest)
        for src, dest in package.get_compilation_files(code)
    )

    download.maybe_add_testlib(code, artifacts)
    download.maybe_add_jngen(code, artifacts)
    download.maybe_add_rbx_header(code, artifacts)
    compilable_path = _maybe_rename_java_class(compilable_path, file_mapping)
    artifacts.inputs.append(
        GradingFileInput(src=compilable_path, dest=PosixPath(file_mapping.compilable))
    )

    artifacts.outputs.append(
        GradingFileOutput(
            src=PosixPath(file_mapping.executable),
            digest=compiled_digest,
            executable=True,
        )
    )

    for input in artifacts.inputs:
        _ignore_warning_in_cxx_input(input)

    if not steps_with_caching.compile(
        commands,
        params=sandbox_params,
        artifacts=artifacts,
        sandbox=sandbox,
        dependency_cache=dependency_cache,
    ):
        raise typer.Exit(1)

    assert compiled_digest.value is not None

    if verbose and artifacts.logs is not None and artifacts.logs.preprocess is not None:
        for log in artifacts.logs.preprocess:
            console.console.print(f'[status]Command:[/status] {log.get_command()}')
            console.console.print(f'[status]Summary:[/status] {log.get_summary()}')
            console.console.print(rich.text.Text.from_ansi(log.log), style='default')

    # Write compiler warnings.
    cfg = setter_config.get_setter_config()
    if (
        (cfg.warnings.enabled or force_warnings)
        and artifacts.logs is not None
        and artifacts.logs.preprocess is not None
    ):
        any_warning = any(log.warnings for log in artifacts.logs.preprocess)
        if any_warning:
            warning_stack.get_warning_stack().add_warning(code)

    # Create sentinel to indicate this executable is sanitized.
    storage = package.get_cache_storage()
    if sanitized.should_sanitize():
        pf = storage.create_file(f'{compiled_digest.value}.san')
        if pf is not None:
            storage.commit_file(pf)
    elif storage.exists(f'{compiled_digest.value}.san'):
        storage.delete(f'{compiled_digest.value}.san')

    return compiled_digest.value


async def run_item(
    code: CodeItem,
    executable: DigestOrSource,
    stdin: Optional[DigestOrSource] = None,
    stdout: Optional[DigestOrDest] = None,
    stderr: Optional[DigestOrDest] = None,
    inputs: Optional[List[GradingFileInput]] = None,
    outputs: Optional[List[GradingFileOutput]] = None,
    extra_args: Optional[str] = None,
    extra_config: Optional[ExecutionConfig] = None,
    retry_index: Optional[int] = None,
) -> Optional[RunLog]:
    _check_stack_limit()

    dependency_cache = package.get_dependency_cache()

    prepared = _prepare_run(
        code,
        executable,
        stdin,
        stdout,
        stderr,
        inputs,
        outputs,
        extra_args,
        extra_config,
        retry_index,
    )

    run_log = await steps_with_caching.run(
        prepared.command,
        params=prepared.sandbox_params,
        sandbox=package.get_singleton_sandbox(),
        artifacts=prepared.artifacts,
        dependency_cache=dependency_cache,
        metadata=prepared.metadata,
    )

    # Find sanitizer logs.
    if run_log is not None and run_log.warnings:
        assert prepared.sandbox_params.stderr_file is not None
        stderr_output = prepared.artifacts.get_output_file_for_src(
            prepared.sandbox_params.stderr_file
        )
        if stderr_output is not None:
            warning_stack.get_warning_stack().add_sanitizer_warning(
                package.get_cache_storage(), code, stderr_output
            )
    return run_log


@dataclasses.dataclass
class CommunicationItem:
    code: CodeItem
    executable: DigestOrSource
    stderr: Optional[DigestOrDest] = None
    inputs: Optional[List[GradingFileInput]] = None
    outputs: Optional[List[GradingFileOutput]] = None
    extra_args: Optional[str] = None
    extra_config: Optional[ExecutionConfig] = None
    capture: Optional[DigestOrDest] = None

    def prepare(self) -> PreparedRun:
        return _prepare_run(
            self.code,
            self.executable,
            stderr=self.stderr,
            inputs=self.inputs,
            outputs=self.outputs,
            extra_args=self.extra_args,
            extra_config=self.extra_config,
        )


async def run_communication(
    interactor: CommunicationItem,
    solution: CommunicationItem,
    merged_capture: Optional[pathlib.Path] = None,
    retry_index: Optional[int] = None,
):
    fifo_in, fifo_out = package.get_fifos()
    interactor_prepared = interactor.prepare()
    solution_prepared = solution.prepare()

    # Prepare retry index.
    interactor_prepared.metadata.retryIndex = retry_index
    solution_prepared.metadata.retryIndex = retry_index

    interactor_prefix = 'INTERACTOR:'
    solution_prefix = 'SOLUTION:'

    if merged_capture is not None:
        package.get_merged_capture_path().write_text(
            f'{interactor_prefix}\n{solution_prefix}\n'
        )

    _prepare_for_communication(
        interactor_prepared,
        fifo_out,
        fifo_in,
        capture=CaptureSpec(
            prefix=interactor_prefix,
            output=interactor.capture,
            merged_capture=merged_capture,
        ),
    )
    _prepare_for_communication(
        solution_prepared,
        fifo_in,
        fifo_out,
        reverse_io=True,
        capture=CaptureSpec(
            prefix=solution_prefix,
            output=solution.capture,
            merged_capture=merged_capture,
        ),
    )

    interactor_run_params = steps.CoordinatedRunParams(
        command=interactor_prepared.command,
        params=interactor_prepared.sandbox_params,
        sandbox=package.get_singleton_interactor_sandbox(),
        artifacts=interactor_prepared.artifacts,
        metadata=interactor_prepared.metadata,
    )
    solution_run_params = steps.CoordinatedRunParams(
        command=solution_prepared.command,
        params=solution_prepared.sandbox_params,
        sandbox=package.get_singleton_sandbox(),
        artifacts=solution_prepared.artifacts,
        metadata=solution_prepared.metadata,
    )

    return await steps_with_caching.run_coordinated(
        interactor_run_params,
        solution_run_params,
        dependency_cache=package.get_dependency_cache(),
    )



import pathlib
from typing import List, Optional, Tuple

from rbx.schema import DumpedProblem


def _normalize_alias(alias: str) -> str:
    return alias.lower()


def _find_alias(alias: str, haystack: List[str]) -> Optional[int]:
    normalized_alias = _normalize_alias(alias)
    for i, candidate in enumerate(haystack):
        if _normalize_alias(candidate) == normalized_alias:
            return i
    return None


def _get_best_alias_from_candidates(
    alias: str, candidates: List[Tuple[pathlib.Path, DumpedProblem]]
) -> Optional[Tuple[pathlib.Path, DumpedProblem]]:
    best_priority = 1e9
    best_candidates = []
    for path, problem in candidates:
        index = _find_alias(alias, problem.aliases)
        if index is None:
            continue
        if index < best_priority:
            best_priority = index
            best_candidates = [(path, problem)]
        elif index == best_priority:
            best_candidates.append((path, problem))

    if len(best_candidates) != 1:
        # TODO
        return None

    return best_candidates[0]


def find_problem_path_by_code(
    code: str, root: Optional[pathlib.Path] = None
) -> Optional[pathlib.Path]:
    if not root:
        root = pathlib.Path()

    metadata_path = root / f'{code}.rbx.json'
    if not metadata_path.is_file():
        return None
    return metadata_path


def find_problem_path_by_alias(
    alias: str, root: Optional[pathlib.Path] = None
) -> Optional[pathlib.Path]:
    if not root:
        root = pathlib.Path()

    candidates: List[Tuple[pathlib.Path, DumpedProblem]] = []
    for metadata_path in root.glob('*.rbx.json'):
        problem = DumpedProblem.model_validate_json(metadata_path.read_text())
        if _find_alias(alias, problem.aliases) is not None:
            candidates.append((metadata_path, problem))

    picked_candidate = _get_best_alias_from_candidates(alias, candidates)
    if not picked_candidate:
        return None
    return picked_candidate[0]


def find_problem_by_alias(
    alias: str, root: Optional[pathlib.Path] = None
) -> Optional[DumpedProblem]:
    metadata_path = find_problem_path_by_alias(alias, root)
    if not metadata_path:
        return None
    return DumpedProblem.model_validate_json(metadata_path.read_text())


def find_problem_by_code(
    code: str, root: Optional[pathlib.Path] = None
) -> Optional[DumpedProblem]:
    metadata_path = find_problem_path_by_code(code, root)
    if not metadata_path:
        return None
    return DumpedProblem.model_validate_json(metadata_path.read_text())


def find_problem_by_anything(
    anything: str, root: Optional[pathlib.Path] = None
) -> Optional[DumpedProblem]:
    problem = find_problem_by_code(anything, root)
    if problem:
        return problem
    return find_problem_by_alias(anything, root)


def find_problems(root: Optional[pathlib.Path] = None) -> List[DumpedProblem]:
    if not root:
        root = pathlib.Path()

    problems = []
    for metadata_path in root.glob('*.rbx.json'):
        problems.append(DumpedProblem.model_validate_json(metadata_path.read_text()))
    return problems



import pathlib
import shutil
from typing import Optional

import typer

from rbx import annotations, console
from rbx.box import header, package
from rbx.box.schema import CodeItem
from rbx.config import get_builtin_checker, get_jngen, get_testlib
from rbx.grading import steps

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


def get_local_artifact(name: str) -> Optional[steps.GradingFileInput]:
    path = pathlib.Path(name)
    if path.is_file():
        return steps.GradingFileInput(src=path, dest=path)
    return None


def maybe_add_rbx_header(code: CodeItem, artifacts: steps.GradingArtifacts):
    header.get_header()
    artifact = get_local_artifact('rbx.h')
    assert artifact is not None
    compilation_files = package.get_compilation_files(code)
    if any(dest == artifact.dest for _, dest in compilation_files):
        return
    artifacts.inputs.append(artifact)


def maybe_add_testlib(code: CodeItem, artifacts: steps.GradingArtifacts):
    # Try to get from compilation files, then from package folder, then from tool.
    artifact = get_local_artifact('testlib.h') or steps.testlib_grading_input()
    compilation_files = package.get_compilation_files(code)
    if any(dest == artifact.dest for _, dest in compilation_files):
        return
    artifacts.inputs.append(artifact)


def maybe_add_jngen(code: CodeItem, artifacts: steps.GradingArtifacts):
    # Try to get from compilation files, then from package folder, then from tool.
    artifact = get_local_artifact('jngen.h') or steps.jngen_grading_input()
    compilation_files = package.get_compilation_files(code)
    if any(dest == artifact.dest for _, dest in compilation_files):
        return
    artifacts.inputs.append(artifact)


@app.command('testlib', help='Download testlib.h')
@package.within_problem
def testlib():
    shutil.copyfile(get_testlib(), pathlib.Path('testlib.h'))
    console.console.print('Downloaded [item]testlib.h[/item] into current package.')


@app.command('jngen', help='Download jngen.h')
@package.within_problem
def jngen():
    shutil.copyfile(get_jngen(), pathlib.Path('jngen.h'))
    console.console.print('Downloaded [item]jngen.h[/item] into current package.')


@app.command('checker', help='Download a built-in checker from testlib GH repo.')
@package.within_problem
def checker(name: str):
    if not name.endswith('.cpp'):
        name = f'{name}.cpp'
    path = get_builtin_checker(name)
    shutil.copyfile(path, pathlib.Path(name))
    console.console.print(
        f'[success]Downloaded [item]{name}[/item] into current package.[/success]'
    )



import pathlib
import tempfile
from typing import Type

import syncer
import typer

from rbx import annotations, console
from rbx.box import environment, header, package
from rbx.box.naming import get_problem_name_with_contest_info
from rbx.box.package import get_build_path
from rbx.box.packaging.packager import BasePackager, BuiltStatement
from rbx.box.statements.build_statements import build_statement

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


async def run_packager(
    packager_cls: Type[BasePackager],
    verification: environment.VerificationParam,
    **kwargs,
) -> pathlib.Path:
    from rbx.box import builder

    header.generate_header()

    if not await builder.verify(verification=verification):
        console.console.print(
            '[error]Build or verification failed, check the report.[/error]'
        )
        raise typer.Exit(1)

    pkg = package.find_problem_package_or_die()

    if pkg.type not in packager_cls.task_types():
        console.console.print(
            f'[error]Packager [item]{packager_cls.name()}[/item] does not support task type [item]{pkg.type}[/item].[/error]'
        )
        raise typer.Exit(1)

    packager = packager_cls(**kwargs)

    statement_types = packager.statement_types()
    built_statements = []

    for statement_type in statement_types:
        languages = packager.languages()
        for language in languages:
            statement = packager.get_statement_for_language(language)
            statement_path = build_statement(statement, pkg, statement_type)
            built_statements.append(
                BuiltStatement(statement, statement_path, statement_type)
            )

    console.console.print(f'Packaging problem for [item]{packager.name()}[/item]...')

    with tempfile.TemporaryDirectory() as td:
        result_path = packager.package(
            get_build_path(), pathlib.Path(td), built_statements
        )

    console.console.print(
        f'[success]Problem packaged for [item]{packager.name()}[/item]![/success]'
    )
    console.console.print(f'Package was saved at [item]{result_path.resolve()}[/item]')
    return result_path


@app.command('polygon', help='Build a package for Polygon.')
@package.within_problem
@syncer.sync
async def polygon(
    verification: environment.VerificationParam,
    upload: bool = typer.Option(
        False,
        '--upload',
        '-u',
        help='If set, will upload the package to Polygon.',
    ),
):
    from rbx.box.packaging.polygon.packager import PolygonPackager

    await run_packager(PolygonPackager, verification=verification)

    if upload:
        from rbx.box.packaging.polygon.upload import upload_problem

        await upload_problem(name=get_problem_name_with_contest_info())


@app.command('boca', help='Build a package for BOCA.')
@package.within_problem
@syncer.sync
async def boca(
    verification: environment.VerificationParam,
    upload: bool = typer.Option(
        False,
        '--upload',
        '-u',
        help='If set, will upload the package to BOCA.',
    ),
):
    from rbx.box.packaging.boca.packager import BocaPackager

    result_path = await run_packager(BocaPackager, verification=verification)

    if upload:
        from rbx.box.packaging.boca.upload import BocaUploader

        uploader = BocaUploader()
        uploader.login_and_upload(result_path)


@app.command('moj', help='Build a package for MOJ.')
@package.within_problem
@syncer.sync
async def moj(
    verification: environment.VerificationParam,
    for_boca: bool = typer.Option(
        False, help='Build a package for BOCA instead of MOJ.'
    ),
):
    from rbx.box.packaging.moj.packager import MojPackager

    await run_packager(MojPackager, verification=verification, for_boca=for_boca)



from typing import Annotated, List, Optional

import syncer
import typer

from rbx import annotations, console
from rbx.box import cd, environment, package
from rbx.box.contest.build_contest_statements import build_statement
from rbx.box.contest.contest_package import (
    find_contest_package_or_die,
    within_contest,
)
from rbx.box.statements.schema import StatementType

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


@app.command('build, b', help='Build statements.')
@within_contest
@syncer.sync
async def build(
    verification: environment.VerificationParam,
    languages: Annotated[
        Optional[List[str]],
        typer.Option(
            default_factory=list,
            help='Languages to build statements for. If not specified, build statements for all available languages.',
        ),
    ],
    output: Annotated[
        Optional[StatementType],
        typer.Option(
            case_sensitive=False,
            help='Output type to be generated. If not specified, will infer from the conversion steps specified in the package.',
        ),
    ] = StatementType.PDF,
    samples: Annotated[
        bool,
        typer.Option(help='Whether to build the statement with samples or not.'),
    ] = True,
    editorial: Annotated[
        bool,
        typer.Option(help='Whether to add editorial blocks to the statements or not.'),
    ] = False,
):
    contest = find_contest_package_or_die()
    # At most run the validators, only in samples.
    if samples:
        from rbx.box import builder

        for problem in contest.problems:
            console.console.print(
                f'Processing problem [item]{problem.short_name}[/item]...'
            )
            with cd.new_package_cd(problem.get_path()):
                package.clear_package_cache()

                if not await builder.build(
                    verification=verification, groups=set(['samples']), output=None
                ):
                    console.console.print(
                        '[error]Failed to build statements with samples, aborting.[/error]'
                    )
                    raise typer.Exit(1)

    contest = find_contest_package_or_die()
    candidate_languages = languages
    if not candidate_languages:
        candidate_languages = sorted(set([st.language for st in contest.statements]))

    for language in candidate_languages:
        candidates_for_lang = [
            st for st in contest.statements if st.language == language
        ]
        if not candidates_for_lang:
            console.console.print(
                f'[error]No contest-level statement found for language [item]{language}[/item].[/error]',
            )
            raise typer.Exit(1)

        build_statement(
            candidates_for_lang[0],
            contest,
            output_type=output,
            use_samples=samples,
            is_editorial=editorial,
        )


@app.callback()
def callback():
    pass



import subprocess
import sys
from pathlib import Path

LAZY_MODULES = {
    'gitpython',
    'questionary',
    'fastapi',
    'requests',
    'pydantic_xml',
    'rbx.box.packaging.polygon.packager',
    'rbx.box.stresses',
}


def test_rich_not_imported_unnecessary():
    file_path = Path(__file__).parent / 'lazy_importing_main.py'
    result = subprocess.run(
        [sys.executable, '-m', 'coverage', 'run', str(file_path)],
        capture_output=True,
        encoding='utf-8',
    )
    modules = result.stdout.splitlines()
    modules = [module for module in modules if module in LAZY_MODULES]
    assert not modules



import functools
import importlib
import importlib.resources
import pathlib
from typing import Callable, Dict, Type

from rbx.box import package
from rbx.box.schema import Primitive


@functools.cache
def get_header() -> pathlib.Path:
    generate_header()

    return pathlib.Path('rbx.h')


def generate_header():
    with importlib.resources.as_file(
        importlib.resources.files('rbx') / 'resources' / 'templates' / 'rbx.h'
    ) as file:
        with file.open('r') as f:
            header = f.read()

    with pathlib.Path('rbx.h').open('w') as f:
        f.write(_preprocess_header(header))


def _preprocess_header(header: str) -> str:
    return (
        header.replace('//<rbx::string_var>', _get_string_var_block())
        .replace('//<rbx::int_var>', _get_int_var_block())
        .replace('//<rbx::float_var>', _get_float_var_block())
        .replace('//<rbx::bool_var>', _get_bool_var_block())
    )


def _get_string_var_block() -> str:
    return _get_var_block(_get_vars_of_type(str, lambda x: f'{x:!r}'))


def _get_int_var_block() -> str:
    return _get_var_block(_get_vars_of_type(int, lambda x: str(x)))


def _get_float_var_block() -> str:
    return _get_var_block(_get_vars_of_type(float, lambda x: f'{x}'))


def _get_bool_var_block() -> str:
    return _get_var_block(_get_vars_of_type(bool, lambda x: 'true' if x else 'false'))


def _get_vars_of_type(
    type: Type, transform: Callable[[Primitive], str]
) -> Dict[str, str]:
    pkg = package.find_problem_package_or_die()
    vars = pkg.expanded_vars
    return {
        name: transform(value)
        for name, value in vars.items()
        if isinstance(value, type)
    }


def _get_var_block(mappings: Dict[str, str]) -> str:
    entries = []
    # Iterate over sorted keys to ensure a deterministic order.
    for name in sorted(mappings):
        value = mappings[name]
        entry = f'  if (name == "{name}") {{\n    return {value};\n  }}\n'
        entries.append(entry)
    return ''.join(entries)



import os
import pathlib
import sys

from rbx.grading import steps_with_caching
from rbx.grading.caching import DependencyCache
from rbx.grading.judge.cacher import FileCacher
from rbx.grading.judge.sandbox import SandboxBase, SandboxParams
from rbx.grading.steps import (
    DigestOrSource,
    GradingArtifacts,
    GradingFileInput,
    GradingFileOutput,
    RunLogMetadata,
)


async def test_run_from_digest(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
    file_cacher: FileCacher,
):
    executable = DigestOrSource.create(file_cacher.put_file_text('print(5)'))
    artifacts = GradingArtifacts()
    artifacts.inputs.append(
        GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
    )
    artifacts.outputs.append(
        GradingFileOutput(src=pathlib.Path('box-out.txt'), dest=pathlib.Path('out.txt'))
    )
    await steps_with_caching.run(
        f'{sys.executable} executable.py',
        params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
        sandbox=sandbox,
        artifacts=artifacts,
        dependency_cache=dependency_cache,
        metadata=RunLogMetadata(),
    )

    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert artifacts.logs.run is not None
    assert artifacts.logs.run.metadata is not None
    assert not artifacts.logs.cached


async def test_run_from_disk(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
):
    pathlib.Path('executable.py').write_text('print(42)')

    executable = DigestOrSource.create(pathlib.Path('executable.py'))
    artifacts = GradingArtifacts()
    artifacts.inputs.append(
        GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
    )
    artifacts.outputs.append(
        GradingFileOutput(src=pathlib.Path('box-out.txt'), dest=pathlib.Path('out.txt'))
    )
    await steps_with_caching.run(
        f'{sys.executable} executable.py',
        params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
        sandbox=sandbox,
        artifacts=artifacts,
        dependency_cache=dependency_cache,
    )

    assert (cleandir / 'out.txt').read_text().strip() == '42'
    assert artifacts.logs is not None
    assert artifacts.logs.run is not None
    assert not artifacts.logs.cached


async def test_run_caches_intermediate_digest_if_dest_changes(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
    file_cacher: FileCacher,
):
    async def configure_and_run_with_dest(dest: pathlib.Path) -> GradingArtifacts:
        executable = DigestOrSource.create(file_cacher.put_file_text('print(5)'))
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(src=pathlib.Path('box-out.txt'), dest=dest)
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached

    another_artifacts = await configure_and_run_with_dest(
        pathlib.Path('another-out.txt')
    )
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert another_artifacts.logs is not None
    assert another_artifacts.logs.cached


async def test_run_overwrite_changed_file_with_storage_value(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
    file_cacher: FileCacher,
):
    async def configure_and_run_with_dest(dest: pathlib.Path) -> GradingArtifacts:
        executable = DigestOrSource.create(file_cacher.put_file_text('print(5)'))
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(src=pathlib.Path('box-out.txt'), dest=dest)
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached

    pathlib.Path('out.txt').write_text('42')

    another_artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert another_artifacts.logs is not None
    assert another_artifacts.logs.cached


async def test_run_recreates_deleted_file_with_storage_value(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
    file_cacher: FileCacher,
):
    async def configure_and_run_with_dest(dest: pathlib.Path) -> GradingArtifacts:
        executable = DigestOrSource.create(file_cacher.put_file_text('print(5)'))
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(src=pathlib.Path('box-out.txt'), dest=dest)
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached

    pathlib.Path('out.txt').unlink()

    another_artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert another_artifacts.logs is not None
    assert another_artifacts.logs.cached


async def test_run_overwrite_exec_bit_when_changed(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
    file_cacher: FileCacher,
):
    async def configure_and_run_with_dest(dest: pathlib.Path) -> GradingArtifacts:
        executable = DigestOrSource.create(file_cacher.put_file_text('print(5)'))
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(
                **executable.expand(),
                dest=pathlib.Path('executable.py'),
            )
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('box-out.txt'), dest=dest, executable=True
            )
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached
    assert os.access('out.txt', os.X_OK)

    pathlib.Path('out.txt').chmod(0o644)

    another_artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert another_artifacts.logs is not None
    assert another_artifacts.logs.cached
    assert os.access('out.txt', os.X_OK)


async def test_run_evicts_when_changed_file_and_no_hash(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
    file_cacher: FileCacher,
):
    async def configure_and_run_with_dest(dest: pathlib.Path) -> GradingArtifacts:
        executable = DigestOrSource.create(file_cacher.put_file_text('print(5)'))
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(src=pathlib.Path('box-out.txt'), dest=dest, hash=False)
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached

    pathlib.Path('out.txt').write_text('42')

    another_artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert another_artifacts.logs is not None
    assert not another_artifacts.logs.cached


async def test_run_evicts_when_exec_bit_different_and_no_hash(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
    file_cacher: FileCacher,
):
    async def configure_and_run_with_dest(dest: pathlib.Path) -> GradingArtifacts:
        executable = DigestOrSource.create(file_cacher.put_file_text('print(5)'))
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('box-out.txt'), dest=dest, hash=False, executable=True
            )
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached

    pathlib.Path('out.txt').chmod(0o0644)

    another_artifacts = await configure_and_run_with_dest(pathlib.Path('out.txt'))
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert another_artifacts.logs is not None
    assert not another_artifacts.logs.cached


async def test_run_evicts_when_input_fingerprint_changes(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
):
    async def configure_and_run() -> GradingArtifacts:
        executable = DigestOrSource.create(pathlib.Path('executable.py'))
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('box-out.txt'),
                dest=pathlib.Path('out.txt'),
            )
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    pathlib.Path('executable.py').write_text('print(5)')

    artifacts = await configure_and_run()
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached

    pathlib.Path('executable.py').write_text('print(42)')

    another_artifacts = await configure_and_run()
    assert (cleandir / 'out.txt').read_text().strip() == '42'
    assert another_artifacts.logs is not None
    assert not another_artifacts.logs.cached


async def test_run_evicts_when_output_is_deleted_and_no_hash(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
):
    async def configure_and_run() -> GradingArtifacts:
        executable = DigestOrSource.create(pathlib.Path('executable.py'))
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('box-out.txt'),
                dest=pathlib.Path('out.txt'),
                hash=False,
            )
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    pathlib.Path('executable.py').write_text('print(5)')

    artifacts = await configure_and_run()
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached

    pathlib.Path('out.txt').unlink()

    another_artifacts = await configure_and_run()
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert another_artifacts.logs is not None
    assert not another_artifacts.logs.cached


async def test_run_misses_when_input_file_changes(
    cleandir: pathlib.Path,
    dependency_cache: DependencyCache,
    sandbox: SandboxBase,
    file_cacher: FileCacher,
):
    async def configure_and_run(number: int) -> GradingArtifacts:
        executable = DigestOrSource.create(
            file_cacher.put_file_text(f'print({number})')
        )
        artifacts = GradingArtifacts()
        artifacts.inputs.append(
            GradingFileInput(**executable.expand(), dest=pathlib.Path('executable.py'))
        )
        artifacts.outputs.append(
            GradingFileOutput(
                src=pathlib.Path('box-out.txt'),
                dest=pathlib.Path('out.txt'),
                hash=False,
            )
        )
        await steps_with_caching.run(
            f'{sys.executable} executable.py',
            params=SandboxParams(stdout_file=pathlib.Path('box-out.txt')),
            sandbox=sandbox,
            artifacts=artifacts,
            dependency_cache=dependency_cache,
        )
        return artifacts

    artifacts = await configure_and_run(5)
    assert (cleandir / 'out.txt').read_text().strip() == '5'
    assert artifacts.logs is not None
    assert not artifacts.logs.cached

    pathlib.Path('out.txt').write_text('42')

    another_artifacts = await configure_and_run(42)
    assert (cleandir / 'out.txt').read_text().strip() == '42'
    assert another_artifacts.logs is not None
    assert not another_artifacts.logs.cached



import dataclasses
import pathlib
import tempfile
import typing
from typing import List, Optional, Tuple

import typer

from rbx import console, testing_utils
from rbx.box import cd, package
from rbx.box.contest.contest_package import get_problems
from rbx.box.contest.schema import Contest, ContestProblem, ContestStatement
from rbx.box.schema import Package, Testcase
from rbx.box.statements import build_statements
from rbx.box.statements.build_statements import (
    get_builders,
    get_environment_languages_for_statement,
    get_relative_assets,
)
from rbx.box.statements.builders import (
    CONTEST_BUILDER_LIST,
    StatementBuilderContest,
    StatementBuilderContext,
    StatementBuilderProblem,
    StatementSample,
    prepare_assets,
)
from rbx.box.statements.joiners import (
    JOINER_LIST,
    StatementJoiner,
    StatementJoinerContext,
)
from rbx.box.statements.schema import Statement, StatementType
from rbx.box.testcase_utils import get_samples


@dataclasses.dataclass
class ExtractedProblem:
    package: Package
    statement: Statement
    problem: ContestProblem
    samples: List[Testcase]
    built_statement: Optional[pathlib.Path] = None

    def get_statement_path(self) -> pathlib.Path:
        return self.problem.get_path() / self.statement.path

    def get_statement_assets(self) -> List[str]:
        return [str(self.problem.get_path() / asset) for asset in self.statement.assets]

    def get_statement_builder_problem(self) -> StatementBuilderProblem:
        return StatementBuilderProblem(
            package=self.package,
            statement=self.statement,
            samples=StatementSample.from_testcases(self.samples),
            io_path=self.built_statement,
            short_name=self.problem.short_name,
        )


def _get_samples(problem: ContestProblem) -> List[Testcase]:
    with cd.new_package_cd(problem.get_path()):
        package.clear_package_cache()
        return get_samples()


def get_statement_builder_problems(
    extracted_problems: List[ExtractedProblem],
) -> List[StatementBuilderProblem]:
    return [ex.get_statement_builder_problem() for ex in extracted_problems]


def get_statement_builder_contest(
    statement: ContestStatement,
    extracted_problems: List[ExtractedProblem],
) -> StatementBuilderContest:
    return StatementBuilderContest(
        title=statement.title,
        location=statement.location,
        date=statement.date,
        problems=get_statement_builder_problems(extracted_problems),
    )


def get_problems_for_statement(
    contest: Contest,
    language: str,
    requires_matching_statement: bool = True,
) -> List[ExtractedProblem]:
    pkgs = get_problems(contest)
    if not pkgs:
        console.console.print(
            '[error]No problems found in the contest, cannot infer statement type.[/error]'
        )
        raise typer.Exit(1)

    res = []
    for pkg, problem in zip(pkgs, contest.problems):
        found = False
        for statement in pkg.statements:
            if statement.language == language or not requires_matching_statement:
                found = True
                res.append(
                    ExtractedProblem(
                        package=pkg,
                        statement=statement,
                        problem=problem,
                        samples=_get_samples(problem),
                    )
                )
                break
        if not found:
            console.console.print(
                f'[error]No statement found for language {language} in problem {problem.short_name}[/error]'
            )
            raise typer.Exit(1)

    return res


def get_builder_problems(
    extracted_problems: List[ExtractedProblem],
) -> List[StatementBuilderProblem]:
    return [
        StatementBuilderProblem(
            package=ex.package,
            statement=ex.statement,
            samples=StatementSample.from_testcases(ex.samples),
        )
        for ex in extracted_problems
    ]


def get_joiner(name: str) -> StatementJoiner:
    for joiner in JOINER_LIST:
        if joiner.name() == name:
            return joiner
    console.console.print(f'[error]Joiner [item]{name}[/item] not found.[/error]')
    raise typer.Exit(1)


def _build_problem_statements(
    statement: ContestStatement,
    contest: Contest,
    root: pathlib.Path,
    output_type: StatementType,
    use_samples: bool = True,
    is_editorial: bool = False,
) -> List[ExtractedProblem]:
    console.console.print('Building problem-level statements...')
    extracted_problems = get_problems_for_statement(contest, statement.language)
    res = []
    contest_cwd_absolute = pathlib.Path().resolve()
    contest_assets = get_relative_assets(statement.path, statement.assets)

    for extracted_problem in extracted_problems:
        console.console.print(
            f'Building statement for problem {extracted_problem.problem.short_name}...'
        )
        with cd.new_package_cd(extracted_problem.problem.get_path()):
            package.clear_package_cache()
            # TODO: respect steps override
            content, _ = build_statements.build_statement_bytes(
                extracted_problem.statement,
                extracted_problem.package,
                output_type=output_type,
                short_name=extracted_problem.problem.short_name,
                overridden_params={
                    cfg.type: cfg for cfg in statement.override.configure
                }
                if statement.override is not None
                else {},  # overridden configure params
                overridden_assets=contest_assets,  # overridden assets
                overridden_params_root=contest_cwd_absolute,
                use_samples=use_samples,
                is_editorial=is_editorial,
            )
        dest_dir = root / '.problems' / extracted_problem.problem.short_name
        dest_path = dest_dir / f'statement{output_type.get_file_suffix()}'
        dest_dir.mkdir(parents=True, exist_ok=True)
        dest_path.write_bytes(content)

        problem_assets = (
            get_relative_assets(
                extracted_problem.get_statement_path(),
                extracted_problem.get_statement_assets(),
            )
            + contest_assets
        )
        prepare_assets(problem_assets, dest_dir)

        res.append(dataclasses.replace(extracted_problem, built_statement=dest_path))
    return res


def build_contest_only(
    statement: ContestStatement,
    contest: Contest,
    extracted_problems: List[ExtractedProblem],
    input: bytes,
    input_type: StatementType,
    output_type: Optional[StatementType] = None,
    is_editorial: bool = False,
) -> Tuple[bytes, StatementType]:
    console.console.print('Building contest-level statement.')
    bdrs = get_builders(
        contest.name,
        statement.steps,
        statement.configure,
        input_type,
        output_type=output_type,
        builder_list=CONTEST_BUILDER_LIST,
    )

    last_content = input
    last_output = input_type
    for bdr, params in bdrs:
        with tempfile.TemporaryDirectory() as td:
            assets = get_relative_assets(
                statement.path, statement.assets
            ) + bdr.inject_assets(pathlib.Path(), params)
            prepare_assets(assets, pathlib.Path(td))
            output = bdr.build(
                input=last_content,
                context=StatementBuilderContext(
                    languages=get_environment_languages_for_statement(),
                    params=params,
                    root=pathlib.Path(td),
                    editorial=is_editorial,
                    vars={**contest.expanded_vars, **statement.expanded_vars},
                ),
                item=get_statement_builder_contest(statement, extracted_problems),
                verbose=False,
            )
        last_content = output
        last_output = bdr.output_type()

    return last_content, last_output


def build_statement_rooted(
    statement: ContestStatement,
    contest: Contest,
    root: pathlib.Path,
    output_type: Optional[StatementType] = None,
    use_samples: bool = True,
    is_editorial: bool = False,
) -> Tuple[bytes, StatementType]:
    # Validate.
    if not statement.path.is_file():
        console.console.print(
            f'[error]Statement file [item]{statement.path}[/item] does not exist for contest.[/error]'
        )
        raise typer.Exit(1)

    if statement.joiner is None:
        joiner = None
        extracted_problems = get_problems_for_statement(
            contest, statement.language, requires_matching_statement=False
        )
    else:
        # Build problem-level statements.
        joiner = get_joiner(statement.joiner.type)
        extracted_problems = _build_problem_statements(
            statement,
            contest,
            root,
            output_type=joiner.joined_type(),
            use_samples=use_samples,
            is_editorial=is_editorial,
        )

    # Build contest-level statement into joiner input type.
    last_content, last_output = build_contest_only(
        statement,
        contest,
        extracted_problems,
        statement.path.read_bytes(),
        statement.type,
        output_type=joiner.joined_type() if joiner is not None else output_type,
        is_editorial=is_editorial,
    )

    if joiner is None:
        return last_content, last_output
    assert statement.joiner is not None

    # Join statements.
    console.console.print('Joining statements...')
    joiner_assets = get_relative_assets(statement.path, statement.assets)
    prepare_assets(joiner_assets, root)

    testing_utils.print_directory_tree(root, show_hidden=True)

    joiner_context = StatementJoinerContext(
        languages=get_environment_languages_for_statement(),
        params=statement.joiner,
        root=root,
    )
    last_content = joiner.build(
        last_content,
        context=joiner_context,
        contest=get_statement_builder_contest(statement, extracted_problems),
    )
    last_output = joiner.output_type()

    # Finish statement.
    last_content, last_output = build_contest_only(
        statement,
        contest,
        extracted_problems,
        last_content,
        last_output,
        output_type=output_type,
        is_editorial=is_editorial,
    )

    return last_content, last_output


def build_statement(
    statement: ContestStatement,
    contest: Contest,
    output_type: Optional[StatementType] = None,
    use_samples: bool = True,
    is_editorial: bool = False,
) -> pathlib.Path:
    with tempfile.TemporaryDirectory() as td:
        root = pathlib.Path(td)
        last_content, last_output = build_statement_rooted(
            statement,
            contest,
            root,
            output_type=output_type,
            use_samples=use_samples,
            is_editorial=is_editorial,
        )

    statement_path = pathlib.Path(
        f'build/{statement.path.stem}{last_output.get_file_suffix()}'
    )
    statement_path.parent.mkdir(parents=True, exist_ok=True)
    statement_path.write_bytes(typing.cast(bytes, last_content))
    console.console.print(
        f'Statement built successfully for language '
        f'[item]{statement.language}[/item] at '
        f'[item]{statement_path.resolve()}[/item]'
    )
    return statement_path



import dataclasses
import pathlib
from abc import ABC, abstractmethod
from typing import List, Tuple

from rbx.box import naming, package
from rbx.box.contest import contest_package
from rbx.box.contest.schema import ContestProblem, ContestStatement
from rbx.box.generators import get_all_built_testcases
from rbx.box.schema import Package, TaskType, Testcase, TestcaseGroup
from rbx.box.statements.schema import Statement, StatementType


@dataclasses.dataclass
class BuiltStatement:
    statement: Statement
    path: pathlib.Path
    output_type: StatementType


@dataclasses.dataclass
class BuiltContestStatement:
    statement: ContestStatement
    path: pathlib.Path
    output_type: StatementType


@dataclasses.dataclass
class BuiltProblemPackage:
    path: pathlib.Path
    package: Package
    problem: ContestProblem


class BasePackager(ABC):
    @classmethod
    @abstractmethod
    def name(cls) -> str:
        pass

    @classmethod
    def task_types(cls) -> List[TaskType]:
        return [TaskType.BATCH]

    def languages(self):
        pkg = package.find_problem_package_or_die()

        res = set()
        for statement in pkg.statements:
            res.add(statement.language)
        return list(res)

    def package_basename(self):
        pkg = package.find_problem_package_or_die()
        shortname = naming.get_problem_shortname()
        if shortname is not None:
            return f'{shortname}-{pkg.name}'
        return pkg.name

    def statement_types(self) -> List[StatementType]:
        return [StatementType.PDF]

    @abstractmethod
    def package(
        self,
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltStatement],
    ) -> pathlib.Path:
        pass

    # Helper methods.
    def get_built_testcases_per_group(self):
        return get_all_built_testcases()

    def get_built_testcases(self) -> List[Tuple[TestcaseGroup, List[Testcase]]]:
        pkg = package.find_problem_package_or_die()
        tests_per_group = self.get_built_testcases_per_group()
        return [(group, tests_per_group[group.name]) for group in pkg.testcases]

    def get_flattened_built_testcases(self) -> List[Testcase]:
        pkg = package.find_problem_package_or_die()
        tests_per_group = self.get_built_testcases_per_group()

        res = []
        for group in pkg.testcases:
            res.extend(tests_per_group[group.name])
        return res

    def get_statement_for_language(self, lang: str) -> Statement:
        pkg = package.find_problem_package_or_die()
        for statement in pkg.statements:
            if statement.language == lang:
                return statement
        raise


class BaseContestPackager(ABC):
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def package(
        self,
        built_packages: List[BuiltProblemPackage],
        build_path: pathlib.Path,
        into_path: pathlib.Path,
        built_statements: List[BuiltContestStatement],
    ) -> pathlib.Path:
        pass

    def languages(self):
        pkg = contest_package.find_contest_package_or_die()

        res = set()
        for statement in pkg.statements:
            res.add(statement.language)
        return list(res)

    def statement_types(self) -> List[StatementType]:
        return [StatementType.PDF]

    def get_statement_for_language(self, lang: str) -> ContestStatement:
        contest = contest_package.find_contest_package_or_die()
        for statement in contest.statements:
            if statement.language == lang:
                return statement
        raise



import datetime
import hashlib
import os
import pathlib
import re
import typing
from typing import Any, NoReturn, Optional, Tuple

import dateparser
import mechanize
import typer
from bs4 import BeautifulSoup

from rbx import console
from rbx.box import naming

ALERT_REGEX = re.compile(r'\<script[^\>]*\>\s*alert\(\'([^\']+)\'\);?\s*\<\/script\>')
UPLOAD_LOG_REGEX = re.compile(r'Problem (\d+) \([^\)]+\) updated')


def _parse_env_var(var: str, override: Optional[str]) -> str:
    if override is not None:
        return override
    value = os.environ.get(var)
    if value is None:
        console.console.print(
            f'[error][item]{var}[/item] is not set. Set it as an environment variable.[/error]'
        )
        raise typer.Exit(1)
    return value


class BocaUploader:
    def __init__(
        self,
        base_url: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ):
        self.base_url = _parse_env_var('BOCA_BASE_URL', base_url)
        self.username = _parse_env_var('BOCA_USERNAME', username)
        self.password = _parse_env_var('BOCA_PASSWORD', password)

        self.br = mechanize.Browser()
        self.br.set_handle_robots(False)
        self.br.addheaders = [  # type: ignore
            (
                'User-agent',
                'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.1) Gecko/2008071615 Fedora/3.0.1-1.fc9 Firefox/3.0.1',
            )
        ]

    def error(self, message: str) -> NoReturn:
        console.console.print(
            f'[error]{message} (at [item]{self.base_url}[/item])[/error]',
        )
        raise typer.Exit(1)

    def raw_error(self, message: str) -> NoReturn:
        console.console.print(f'[error]{message}[/error]')
        raise typer.Exit(1)

    def log_response_alert(self, response: Any, message: str) -> Tuple[Any, str]:
        if response is None:
            self.raw_error(
                f'{message} ([item]{self.base_url}[/item]):\nNo response received.'
            )
        html = response.read().decode()
        alert = ALERT_REGEX.search(html)
        if alert:
            self.raw_error(
                f'{message} ([item]{self.base_url}[/item]):\n{alert.group(1)}'
            )
        return response, html

    def check_logs_for_update(self, problem_id: int) -> bool:
        _, html = self.open(
            f'{self.base_url}/admin/log.php',
            error_msg='Error while checking whether package upload was successful',
        )

        soup = BeautifulSoup(html, 'html.parser')
        table = soup.select_one('table:nth-of-type(3)')
        if table is None:
            self.raw_error(
                'Error while checking whether package upload was successful:\nNo logs table found.'
            )
        rows = table.select('tr:not(:first-child)')
        for row in rows:
            # Check whether the log line is recent.
            date_cell = row.select('td:nth-of-type(5)')
            if date_cell is None:
                continue
            date = date_cell[0].text.strip()

            parsed_date = dateparser.parse(date)
            if parsed_date is None:
                console.console.print(
                    f'Error while checking whether package upload was successful:\nCould not parse date [item]{date}[/item].'
                )
                raise typer.Exit(1)
            if parsed_date < datetime.datetime.now(
                datetime.timezone.utc
            ) - datetime.timedelta(minutes=1):
                continue

            # Check if the log line contains the problem id.
            log_cell = row.select('td:nth-of-type(6)')
            if log_cell is None:
                continue
            log_line = log_cell[0].text.strip()
            match = UPLOAD_LOG_REGEX.match(log_line)
            if match is None:
                continue
            found_id = int(match.group(1))
            if found_id == problem_id:
                return True
        return False

    def check_submit(self, response: Any, problem_id: int) -> bool:
        if response is None:
            self.raw_error(
                'Error while submitting problem to BOCA website:\nNo response received.'
            )
        html = response.read().decode()
        alert = ALERT_REGEX.search(html)
        if alert:
            msg = alert.group(1)
            if 'Violation' in msg:
                return False
            self.raw_error(
                f'Error while submitting problem to BOCA website:\n{alert.group(1)}'
            )
        return self.check_logs_for_update(problem_id)

    def open(self, url: str, error_msg: Optional[str] = None):
        if error_msg is None:
            error_msg = f'Error while opening [item]{url}[/item]'
        response = self.br.open(url)
        return self.log_response_alert(response, error_msg)

    def login(self):
        _, html = self.open(
            f'{self.base_url}', error_msg='Error while opening BOCA login page'
        )

        needle = "js_myhash(document.form1.password.value)+'"
        start = html.index(needle)
        start_salt = start + len(needle)
        end_salt = html.index("'", start_salt)
        salt = html[start_salt:end_salt]
        console.console.print(f'Using salt [item]{salt}[/item]')

        pwd_hash = hashlib.sha256(self.password.encode()).hexdigest()
        pwd_hash = hashlib.sha256((pwd_hash + salt).encode()).hexdigest()

        login_url = f'{self.base_url}?name={self.username}&password={pwd_hash}'
        self.open(login_url, error_msg='Error while logging in to BOCA')

    def upload(self, file: pathlib.Path) -> bool:
        self.open(
            f'{self.base_url}/admin/problem.php',
            error_msg='Error while opening BOCA problem upload page',
        )
        try:
            form = self.br.select_form(name='form1')
        except mechanize.FormNotFoundError:
            self.error(
                'Problem upload form not found in BOCA website. This might happen when the login failed.'
            )

        form = typing.cast(mechanize.HTMLForm, self.br.form)
        form.set_all_readonly(False)

        problem_index = naming.get_problem_index()
        if problem_index is None:
            console.console.print(
                'It seems this problem is not part of a contest. Cannot upload it to BOCA.'
            )
            raise typer.Exit(1)

        problem_shortname = naming.get_problem_shortname()
        assert problem_shortname is not None
        entry = naming.get_problem_entry_in_contest()
        assert entry is not None
        _, problem_entry = entry

        hex_color = problem_entry.hex_color
        if hex_color is None:
            form['colorname'] = 'black'
            form['color'] = '000000'
        else:
            assert problem_entry.color_name is not None
            form['colorname'] = problem_entry.color_name.lower()
            form['color'] = hex_color[1:]

        form['problemnumber'] = f'{problem_index + 1}'
        form['problemname'] = problem_shortname
        form['confirmation'] = 'confirm'
        form['autojudge_new_sel'] = ['all']
        form['Submit3'] = 'Send'

        with file.open('rb') as f:
            form.add_file(
                f,
                filename=file.name,
                name='probleminput',
                content_type='application/zip',
            )
            response = self.br.submit()

        return self.check_submit(response, problem_index + 1)

    def login_and_upload(self, file: pathlib.Path):
        RETRIES = 3

        tries = 0
        ok = False
        while tries < RETRIES:
            tries += 1

            console.console.print('Logging in to BOCA...')
            self.login()
            console.console.print('Uploading problem to BOCA...')
            if not self.upload(file):
                console.console.print(
                    f'[warning]Potentially transient error while uploading problem to BOCA. Retrying ({tries}/{RETRIES})...[/warning]'
                )
                continue

            ok = True
            console.console.print(
                '[success]Problem sent to BOCA. [item]rbx[/item] cannot determine the upload succeeded, check the website to be sure.[/success]'
            )
            break

        if not ok:
            console.console.print(
                '[error]Persistent error while uploading problem to BOCA website.[/error]'
            )
            console.console.print(
                '[warning]This might be caused by PHP max upload size limit (which usually defaults to 2MBF).[/warning]'
            )
            console.console.print(
                '[warning]Check [item]https://www.php.net/manual/en/ini.core.php#ini.sect.file-uploads[/item] for more information.[/warning]'
            )
            raise typer.Exit(1)



import asyncio
import contextlib
import os
import signal
import threading
from typing import Optional, Set

_processing_context_pids: Optional[Set[int]] = None
_terminate_all_on_error = False
_lock = threading.Lock()

# Creating a processing context is not thread-safe, but adding to it is.


@contextlib.contextmanager
def new_processing_context(terminate_all_on_error: bool = False):
    global _processing_context_pids, _terminate_all_on_error
    with _lock:
        old_processing_context_pids = _processing_context_pids
        _old_terminate_all_on_error = _terminate_all_on_error
        _processing_context_pids = set()
        _terminate_all_on_error = terminate_all_on_error
    try:
        yield
    finally:
        with _lock:
            _processing_context_pids = old_processing_context_pids
            _terminate_all_on_error = _old_terminate_all_on_error


def get_processing_context() -> Set[int]:
    with _lock:
        return _processing_context_pids or set()


def add_to_processing_context(pid: int):
    global _processing_context_pids
    with _lock:
        if _processing_context_pids is None:
            return
        _processing_context_pids.add(pid)


def terminate_all_processes_in_context(clear: bool = True):
    global _processing_context_pids
    with _lock:
        if _processing_context_pids is None:
            return
        for pid in _processing_context_pids:
            try:
                os.kill(pid, signal.SIGTERM)
            except OSError:
                pass
        if clear:
            _processing_context_pids.clear()


async def wait_all_processes_in_context(wait_for: int):
    global _processing_context_pids, _terminate_all_on_error
    wait_pids = set()
    while len(get_processing_context()) < wait_for:
        await asyncio.sleep(0.01)

    with _lock:
        if _processing_context_pids is None:
            return
        wait_pids.update(_processing_context_pids)

    wait_lock = threading.Lock()
    finished_pids = []

    def process(pid: int, returncode: int):
        with wait_lock:
            finished_pids.append(pid)
        if returncode != 0 and _terminate_all_on_error:
            terminate_all_processes_in_context()

    def wait_all_processes():
        while len(finished_pids) < len(wait_pids):
            try:
                pid, status = os.wait()
            except ChildProcessError:
                return
            if pid in wait_pids:
                process(pid, os.waitstatus_to_exitcode(status))

    await asyncio.to_thread(wait_all_processes)



from typing import Awaitable, Callable, Generic, Optional, TypeVar

T = TypeVar('T')
U = TypeVar('U')


class Deferred(Generic[T]):
    def __init__(self, func: Callable[[], Awaitable[T]]):
        self.func = func
        self.cache: Optional[T] = None

    def __call__(self) -> Awaitable[T]:
        async def async_wrapper():
            if self.cache is None:
                self.cache = await self.func()
            return self.cache

        return async_wrapper()

    def peek(self) -> Optional[T]:
        return self.cache

    def wrap_with(
        self, wrapper: Callable[[Awaitable[T]], Awaitable[U]]
    ) -> 'Deferred[U]':
        return Deferred(lambda: wrapper(self()))



import pathlib
from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, model_validator

from rbx.box.schema import NameField, Primitive, expand_var
from rbx.box.statements.schema import (
    ConversionStep,
    Joiner,
    StatementType,
)


def ShortNameField(**kwargs):
    return Field(pattern=r'^[A-Z]+[0-9]*$', min_length=1, max_length=4, **kwargs)


class ProblemStatementOverride(BaseModel):
    model_config = ConfigDict(extra='forbid')

    configure: List[ConversionStep] = Field(
        default=[],
        discriminator='type',
        description="""
Configure how certain conversion steps should happen when applied to the statement file.

Different from the `steps` field, this does not force the steps to happen, but rather only
configure them in case they are applied.
""",
    )


class ContestStatement(BaseModel):
    model_config = ConfigDict(extra='forbid')

    language: str = Field(default='en', description='Language code for this statement.')

    title: str = Field(description='Title of the contest in this language.')

    location: Optional[str] = Field(
        default=None, description='Location of the contest in this language.'
    )

    date: Optional[str] = Field(
        default=None, description='Date of the contest in this language.'
    )

    path: pathlib.Path = Field(description='Path to the input statement file.')

    type: StatementType = Field(description='Type of the input statement file.')

    joiner: Optional[Joiner] = Field(
        default=None,
        description="""
Joiner to be used to build the statement.
                           
This determines how problem statements will be joined into a single contest statement.""",
    )

    steps: List[ConversionStep] = Field(
        default=[],
        discriminator='type',
        description="""
Describes a sequence of conversion steps that should be applied to the statement file
of this contest.

Usually, it is not necessary to specify these, as they can be inferred from the
input statement type and the output statement type, but you can use this to force
certain conversion steps to happen.
""",
    )

    configure: List[ConversionStep] = Field(
        default=[],
        discriminator='type',
        description="""
Configure how certain conversion steps should happen when applied to the statement file of
this contest.

Different from the `steps` field, this does not force the steps to happen, but rather only
configure them in case they are applied.
""",
    )

    assets: List[str] = Field(
        default=[],
        description="""
Assets relative to the contest directory that should be included while building
the statement. Files will be included in the same folder as the statement file.
Can be glob pattern as well, such as `imgs/*.png`.
""",
    )

    override: Optional[ProblemStatementOverride] = Field(
        default=None, description='Override configuration for problem statements.'
    )

    # Vars to be re-used in the statement.
    #   - It will be available as \VAR{vars} variable in the contest-level box statement.
    vars: Dict[str, Primitive] = Field(
        default={}, description='Variables to be re-used across the package.'
    )

    @property
    def expanded_vars(self) -> Dict[str, Primitive]:
        return {key: expand_var(value) for key, value in self.vars.items()}


class ContestProblem(BaseModel):
    short_name: str = ShortNameField(
        description="""
Short name of the problem. Usually, just an uppercase letter,
but can be a sequence of uppercase letters followed by a number."""
    )
    path: Optional[pathlib.Path] = Field(
        default=None,
        description="""
Path to the problem relative to the contest package directory.
If not specified, will expect the problem to be in ./{short_name}/ folder.""",
    )

    color: Optional[str] = Field(
        default=None,
        description="""
Color that represents this problem in the contest.

Can be a hex color (#abcdef or #abc format), or a color name among available X11 colors.

See https://en.wikipedia.org/wiki/X11_color_names for the list of supported color names.
""",
    )

    colorName: Optional[str] = Field(
        default=None,
        description="""
A custom color name for the color provided by this problem.

If not provided, will try to infer a color name from the color provided.
""",
        pattern=r'^[a-zA-Z]+$',
    )

    @model_validator(mode='after')
    def check_color(self):
        from colour import Color

        if self.color is None:
            return self

        Color(self.color)
        return self

    @property
    def hex_color(self) -> Optional[str]:
        from colour import Color

        if self.color is None:
            return None

        return Color(self.color).hex_l

    @property
    def color_name(self) -> Optional[str]:
        if self.colorName is not None:
            return self.colorName

        if self.color is None:
            return None

        from colour import Color

        color = Color(self.color)
        web_color = color.web
        if web_color.startswith('#'):
            return 'unknown'
        return web_color

    def get_path(self) -> pathlib.Path:
        return self.path or pathlib.Path(self.short_name)


class Contest(BaseModel):
    model_config = ConfigDict(extra='forbid')

    name: str = NameField(description='Name of this contest.')

    problems: List[ContestProblem] = Field(
        default=[], description='List of problems in this contest.'
    )

    statements: List[ContestStatement] = Field(
        default=None,
        description='Configure statements in this contest, per language.',
    )

    # Vars to be re-used in the statements.
    #   - It will be available as \VAR{vars} variable in the contest-level box statement.
    vars: Dict[str, Primitive] = Field(
        default={}, description='Variables to be re-used across the package.'
    )

    @property
    def expanded_vars(self) -> Dict[str, Primitive]:
        return {key: expand_var(value) for key, value in self.vars.items()}



import pathlib
import tempfile
from typing import Type

import syncer
import typer

from rbx import annotations, console
from rbx.box import cd, environment, package
from rbx.box.contest import build_contest_statements, contest_package
from rbx.box.packaging.main import run_packager
from rbx.box.packaging.packager import (
    BaseContestPackager,
    BasePackager,
    BuiltContestStatement,
    BuiltProblemPackage,
)

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


async def run_contest_packager(
    contest_packager_cls: Type[BaseContestPackager],
    packager_cls: Type[BasePackager],
    verification: environment.VerificationParam,
):
    contest = contest_package.find_contest_package_or_die()

    # Build problem-level packages.
    built_packages = []
    for problem in contest.problems:
        console.console.print(
            f'Processing problem [item]{problem.short_name}[/item]...'
        )
        with cd.new_package_cd(problem.get_path()):
            package.clear_package_cache()
            package_path = await run_packager(packager_cls, verification=verification)
            built_packages.append(
                BuiltProblemPackage(
                    path=problem.get_path() / package_path,
                    package=package.find_problem_package_or_die(),
                    problem=problem,
                )
            )

    # Build statements.
    packager = contest_packager_cls()
    statement_types = packager.statement_types()
    built_statements = []

    for statement_type in statement_types:
        languages = packager.languages()
        for language in languages:
            statement = packager.get_statement_for_language(language)
            statement_path = build_contest_statements.build_statement(
                statement, contest, statement_type
            )
            built_statements.append(
                BuiltContestStatement(statement, statement_path, statement_type)
            )

    console.console.print(f'Packaging contest for [item]{packager.name()}[/item]...')

    # Build contest-level package.
    with tempfile.TemporaryDirectory() as td:
        packager.package(
            built_packages, pathlib.Path('build'), pathlib.Path(td), built_statements
        )

    console.console.print(
        f'[success]Contest packaged for [item]{packager.name()}[/item]![/success]'
    )


@app.command('polygon', help='Build a contest package for Polygon.')
@contest_package.within_contest
@syncer.sync
async def polygon(
    verification: environment.VerificationParam,
):
    from rbx.box.packaging.polygon.packager import (
        PolygonContestPackager,
        PolygonPackager,
    )

    await run_contest_packager(
        PolygonContestPackager, PolygonPackager, verification=verification
    )



import pathlib
from typing import Optional

import git


def get_repo_or_nil(
    root: pathlib.Path = pathlib.Path(), search_parent_directories: bool = False
) -> Optional[git.Repo]:
    try:
        return git.Repo(root, search_parent_directories=search_parent_directories)
    except git.InvalidGitRepositoryError:
        return None


def is_repo(path: pathlib.Path) -> bool:
    return get_repo_or_nil(path, search_parent_directories=False) is not None


def is_within_repo(path: pathlib.Path) -> bool:
    return get_repo_or_nil(path, search_parent_directories=True) is not None


def get_any_remote(repo: git.Repo) -> Optional[git.Remote]:
    for remote in repo.remotes:
        if remote.exists():
            return remote
    return None



import pathlib
import tempfile
import typing
from typing import Annotated, Dict, List, Optional, Tuple

import syncer
import typer

from rbx import annotations, console
from rbx.box import environment, naming, package
from rbx.box.schema import Package
from rbx.box.statements.builders import (
    BUILDER_LIST,
    PROBLEM_BUILDER_LIST,
    StatementBuilder,
    StatementBuilderContext,
    StatementBuilderProblem,
    StatementCodeLanguage,
    StatementSample,
    prepare_assets,
)
from rbx.box.statements.schema import (
    ConversionStep,
    ConversionType,
    Statement,
    StatementType,
)
from rbx.box.testcase_utils import get_samples

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


def get_environment_languages_for_statement() -> List[StatementCodeLanguage]:
    env = environment.get_environment()

    res = []
    for language in env.languages:
        cmd = ''
        compilation_cfg = environment.get_compilation_config(language.name)
        cmd = ' && '.join(compilation_cfg.commands or [])
        if not cmd:
            execution_cfg = environment.get_execution_config(language.name)
            cmd = execution_cfg.command

        res.append(
            StatementCodeLanguage(
                id=language.name,
                name=language.readable_name or language.name,
                command=cmd or '',
            )
        )

    return res


def get_builder(
    name: ConversionType, builder_list: List[StatementBuilder]
) -> StatementBuilder:
    candidates = [builder for builder in builder_list if builder.name() == name]
    if not candidates:
        console.console.print(
            f'[error]No statement builder found with name [name]{name}[/name][/error]'
        )
        raise typer.Exit(1)
    return candidates[0]


def get_implicit_builders(
    input_type: StatementType, output_type: StatementType
) -> Optional[List[StatementBuilder]]:
    par: Dict[StatementType, Optional[StatementBuilder]] = {input_type: None}

    def _iterate() -> bool:
        nonlocal par
        for bdr in BUILDER_LIST:
            u = bdr.input_type()
            if u not in par:
                continue
            v = bdr.output_type()
            if v in par:
                continue
            par[v] = bdr
            return True
        return False

    while _iterate() and output_type not in par:
        pass

    if output_type not in par:
        return None

    res = []
    cur = output_type
    while par[cur] is not None:
        res.append(par[cur])
        cur = typing.cast(StatementBuilder, par[cur]).input_type()

    return list(reversed(res))


def _try_implicit_builders(
    statement_id: str, input_type: StatementType, output_type: StatementType
) -> List[StatementBuilder]:
    implicit_builders = get_implicit_builders(input_type, output_type)
    if implicit_builders is None:
        console.console.print(
            f'[error]Cannot implicitly convert statement [item]{statement_id}[/item] '
            f'from [item]{input_type}[/item] '
            f'to specified output type [item]{output_type}[/item].[/error]'
        )
        raise typer.Exit(1)
    console.console.print(
        'Implicitly adding statement builders to convert statement '
        f'from [item]{input_type}[/item] to [item]{output_type}[/item]...'
    )
    return implicit_builders


def _get_configured_params_for(
    configure: List[ConversionStep], conversion_type: ConversionType
) -> Optional[ConversionStep]:
    for step in configure:
        if step.type == conversion_type:
            return step
    return None


def get_builders(
    statement_id: str,
    steps: List[ConversionStep],
    configure: List[ConversionStep],
    input_type: StatementType,
    output_type: Optional[StatementType],
    builder_list: List[StatementBuilder] = BUILDER_LIST,
) -> List[Tuple[StatementBuilder, ConversionStep]]:
    last_output = input_type
    builders: List[Tuple[StatementBuilder, ConversionStep]] = []

    # Conversion steps to force during build.
    for step in steps:
        builder = get_builder(step.type, builder_list=builder_list)
        if builder.input_type() != last_output:
            implicit_builders = _try_implicit_builders(
                statement_id, last_output, builder.input_type()
            )
            builders.extend(
                (builder, builder.default_params()) for builder in implicit_builders
            )
        builders.append((builder, step))
        last_output = builder.output_type()

    if output_type is not None and last_output != output_type:
        implicit_builders = _try_implicit_builders(
            statement_id, last_output, output_type
        )
        builders.extend(
            (builder, builder.default_params()) for builder in implicit_builders
        )

    # Override statement configs.
    def reconfigure(params: ConversionStep) -> ConversionStep:
        new_params = _get_configured_params_for(configure, params.type)
        return new_params or params

    reconfigured_builders = [
        (builder, reconfigure(params)) for builder, params in builders
    ]
    return reconfigured_builders


def get_relative_assets(
    relative_to: pathlib.Path,
    assets: List[str],
) -> List[Tuple[pathlib.Path, pathlib.Path]]:
    relative_to = relative_to.resolve()
    if not relative_to.is_dir():
        relative_to = relative_to.parent
    res = []
    for asset in assets:
        relative_path = pathlib.Path(asset)
        if not relative_path.is_file():
            globbed = list(
                path
                for path in pathlib.Path().glob(str(relative_path))
                if path.is_file()
            )
            if not globbed and '*' not in str(relative_path):
                console.console.print(
                    f'[error]Asset [item]{asset}[/item] does not exist.[/error]'
                )
                raise typer.Exit(1)
            res.extend(get_relative_assets(relative_to, list(map(str, globbed))))
            continue
        if not relative_path.resolve().is_relative_to(relative_to):
            console.console.print(
                f'[error]Asset [item]{asset}[/item] is not relative to your statement.[/error]'
            )
            raise typer.Exit(1)

        res.append(
            (
                relative_path.resolve(),
                relative_path.resolve().relative_to(relative_to),
            )
        )

    return res


def build_statement_bytes(
    statement: Statement,
    pkg: Package,
    output_type: Optional[StatementType] = None,
    short_name: Optional[str] = None,
    overridden_params_root: pathlib.Path = pathlib.Path(),
    overridden_params: Optional[Dict[ConversionType, ConversionStep]] = None,
    overridden_assets: Optional[List[Tuple[pathlib.Path, pathlib.Path]]] = None,
    use_samples: bool = True,
    is_editorial: bool = False,
) -> Tuple[bytes, StatementType]:
    overridden_params = overridden_params or {}
    overridden_assets = overridden_assets or []

    if not statement.path.is_file():
        console.console.print(
            f'[error]Statement file [item]{statement.path}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)
    builders = get_builders(
        str(statement.path),
        statement.steps,
        statement.configure,
        statement.type,
        output_type,
        builder_list=PROBLEM_BUILDER_LIST,
    )
    last_output = statement.type
    last_content = statement.path.read_bytes()
    for bdr, params in builders:
        with tempfile.TemporaryDirectory() as td:
            # Here, create a new temp context for each builder call.
            assets = get_relative_assets(statement.path, statement.assets)

            # Use either overridden assets (by contest) or usual assets.
            # Remember to modify the root to contest root if that's the case.
            if bdr.name() in overridden_params:
                assets.extend(
                    bdr.inject_assets(
                        overridden_params_root, overridden_params[bdr.name()]
                    )
                )
            else:
                assets.extend(bdr.inject_assets(pathlib.Path(), params))
            assets.extend(overridden_assets)

            prepare_assets(assets, pathlib.Path(td))
            output = bdr.build(
                input=last_content,
                context=StatementBuilderContext(
                    languages=get_environment_languages_for_statement(),
                    params=params,
                    root=pathlib.Path(td),
                    editorial=is_editorial,
                ),
                item=StatementBuilderProblem(
                    package=pkg,
                    statement=statement,
                    samples=StatementSample.from_testcases(
                        get_samples() if use_samples else []
                    ),
                    short_name=short_name,
                ),
                verbose=False,
            )
        last_output = bdr.output_type()
        last_content = output

    return last_content, last_output


def build_statement(
    statement: Statement,
    pkg: Package,
    output_type: Optional[StatementType] = None,
    use_samples: bool = True,
    is_editorial: bool = False,
) -> pathlib.Path:
    last_content, last_output = build_statement_bytes(
        statement,
        pkg,
        output_type=output_type,
        use_samples=use_samples,
        is_editorial=is_editorial,
        short_name=naming.get_problem_shortname(),
    )
    statement_path = (
        package.get_build_path()
        / f'{statement.path.stem}{last_output.get_file_suffix()}'
    )
    statement_path.parent.mkdir(parents=True, exist_ok=True)
    statement_path.write_bytes(last_content)
    console.console.print(
        f'Statement built successfully for language '
        f'[item]{statement.language}[/item] at '
        f'[item]{statement_path.resolve()}[/item]'
    )
    return statement_path


@app.command('build, b', help='Build statements.')
@package.within_problem
@syncer.sync
async def build(
    verification: environment.VerificationParam,
    languages: Annotated[
        Optional[List[str]],
        typer.Option(
            default_factory=list,
            help='Languages to build statements for. If not specified, build statements for all available languages.',
        ),
    ],
    output: Annotated[
        Optional[StatementType],
        typer.Option(
            case_sensitive=False,
            help='Output type to be generated. If not specified, will infer from the conversion steps specified in the package.',
        ),
    ] = StatementType.PDF,
    samples: Annotated[
        bool,
        typer.Option(help='Whether to build the statement with samples or not.'),
    ] = True,
    editorial: Annotated[
        bool, typer.Option(help='Whether to add editorial blocks to the statements.')
    ] = False,
):
    # At most run the validators, only in samples.
    if samples:
        from rbx.box import builder

        if not await builder.build(
            verification=verification,
            groups=set(['samples']),
            output=None,
        ):
            console.console.print(
                '[error]Failed to build statements with samples, aborting.[/error]'
            )
            raise typer.Exit(1)

    pkg = package.find_problem_package_or_die()
    candidate_languages = languages
    if not candidate_languages:
        candidate_languages = sorted(set([st.language for st in pkg.statements]))

    for language in candidate_languages:
        candidates_for_lang = [st for st in pkg.statements if st.language == language]
        if not candidates_for_lang:
            console.console.print(
                f'[error]No statement found for language [item]{language}[/item].[/error]',
            )
            raise typer.Exit(1)

        build_statement(
            candidates_for_lang[0],
            pkg,
            output_type=output,
            use_samples=samples,
            is_editorial=editorial,
        )


@app.callback()
def callback():
    pass



import dataclasses
import io
import logging
import pathlib
import tempfile
from abc import ABC, abstractmethod
from typing import IO, AnyStr, List, Optional

logger = logging.getLogger(__name__)

TOMBSTONE = 'x'


def copyfileobj(
    source_fobj: IO[AnyStr],
    destination_fobj: IO[AnyStr],
    buffer_size=io.DEFAULT_BUFFER_SIZE,
    maxlen: Optional[int] = None,
):
    """Read all content from one file object and write it to another.
    Repeatedly read from the given source file object, until no content
    is left, and at the same time write the content to the destination
    file object. Never read or write more than the given buffer size.
    Be cooperative with other greenlets by yielding often.
    source_fobj (fileobj): a file object open for reading, in either
        binary or str mode (doesn't need to be buffered).
    destination_fobj (fileobj): a file object open for writing, in the
        same mode as the source (doesn't need to be buffered).
    buffer_size (int): the size of the read/write buffer.
    maxlen (int): the maximum number of bytes to copy. If None, copy all.
    """
    if maxlen is None:
        maxlen = -1
    while maxlen:
        buffer = source_fobj.read(buffer_size)
        if len(buffer) == 0:
            break
        if maxlen > 0 and maxlen < len(buffer):
            buffer = buffer[:maxlen]
        while len(buffer) > 0:
            written = destination_fobj.write(buffer)
            buffer = buffer[written:]
            maxlen -= written


@dataclasses.dataclass
class PendingFile:
    fd: IO[bytes]
    filename: str


@dataclasses.dataclass
class FileWithDescription:
    filename: str
    description: str


class Storage(ABC):
    """Abstract base class for all concrete storages."""

    @abstractmethod
    def get_file(self, filename: str) -> IO[bytes]:
        """Retrieve a file from the storage.
        filename (unicode): the path of the file to retrieve.
        return (fileobj): a readable binary file-like object from which
            to read the contents of the file.
        raise (KeyError): if the file cannot be found.
        """
        pass

    @abstractmethod
    def create_file(self, filename: str) -> Optional[PendingFile]:
        """Create an empty file that will live in the storage.
        Once the caller has written the contents to the file, the commit_file()
        method must be called to commit it into the store.
        filename (unicode): the filename of the file to store.
        return (fileobj): a writable binary file-like object on which
            to write the contents of the file, or None if the file is
            already stored.
        """
        pass

    @abstractmethod
    def commit_file(self, file: PendingFile, desc: str = '') -> bool:
        """Commit a file created by create_file() to be stored.
        Given a file object returned by create_file(), this function populates
        the database to record that this file now legitimately exists and can
        be used.
        fobj (fileobj): the object returned by create_file()
        file (PendingFile): the file to commit.
        return (bool): True if the file was committed successfully, False if
            there was already a file with the same filename in the database. This
            shouldn't make any difference to the caller, except for testing
            purposes!
        """
        pass

    @abstractmethod
    def exists(self, filename: str) -> bool:
        """Check if a file exists in the storage."""
        pass

    @abstractmethod
    def describe(self, filename: str) -> str:
        """Return the description of a file given its filename.
        filename (unicode): the filename of the file to describe.
        return (unicode): the description of the file.
        raise (KeyError): if the file cannot be found.
        """
        pass

    @abstractmethod
    def get_size(self, filename: str) -> int:
        """Return the size of a file given its filename.
        filename (unicode): the filename of the file to calculate the size
            of.
        return (int): the size of the file, in bytes.
        raise (KeyError): if the file cannot be found.
        """
        pass

    @abstractmethod
    def delete(self, filename: str):
        """Delete a file from the storage.
        filename (unicode): the filename of the file to delete.
        """
        pass

    @abstractmethod
    def list(self) -> List[FileWithDescription]:
        """List the files available in the storage.
        return ([(unicode, unicode)]): a list of pairs, each
            representing a file in the form (filename, description).
        """
        pass

    @abstractmethod
    def path_for_symlink(self, filename: str) -> Optional[pathlib.Path]:
        pass


class NullStorage(Storage):
    """This backend is always empty, it just drops each file that
    receives. It looks mostly like /dev/null. It is useful when you
    want to just rely on the caching capabilities of FileCacher for
    very short-lived and local storages.

    """

    def get_file(self, digest: str) -> IO[bytes]:
        raise KeyError('File not found.')

    def create_file(self, digest: str) -> Optional[PendingFile]:
        return None

    def commit_file(self, file: PendingFile, desc: str = '') -> bool:
        return False

    def exists(self, filename: str) -> bool:
        return False

    def describe(self, digest: str) -> str:
        raise KeyError('File not found.')

    def get_size(self, digest: str) -> int:
        raise KeyError('File not found.')

    def delete(self, digest: str):
        pass

    def list(self) -> List[FileWithDescription]:
        return list()

    def path_for_symlink(self, digest: str) -> Optional[pathlib.Path]:
        return None


class FilesystemStorage(Storage):
    """This class implements a backend for FileCacher that keeps all
    the files in a file system directory, named after their filename.
    """

    def __init__(self, path: pathlib.Path):
        """Initialize the backend.
        path (string): the base path for the storage.
        """
        self.path = path

        # Create the directory if it doesn't exist
        path.mkdir(parents=True, exist_ok=True)

    def get_file(self, filename: str) -> IO[bytes]:
        """See FileCacherBackend.get_file()."""
        file_path = self.path / filename

        if not file_path.is_file():
            raise KeyError('File not found.')

        return file_path.open('rb')

    def create_file(self, filename: str) -> Optional[PendingFile]:
        """See FileCacherBackend.create_file()."""
        # Check if the file already exists. Return None if so, to inform the
        # caller they don't need to store the file.
        file_path = self.path / filename

        if file_path.is_file():
            return None

        # Create a temporary file in the same directory
        temp_file = tempfile.NamedTemporaryFile(
            'wb', delete=False, prefix='.tmp.', suffix=filename, dir=self.path
        )
        return PendingFile(fd=temp_file, filename=filename)

    def commit_file(self, file: PendingFile, desc: str = '') -> bool:
        """See FileCacherBackend.commit_file()."""
        file.fd.close()

        file_path: pathlib.Path = self.path / file.filename
        # Move it into place in the cache. Skip if it already exists, and
        # delete the temporary file instead.
        if not file_path.is_file():
            # There is a race condition here if someone else puts the file here
            # between checking and renaming. Put it doesn't matter in practice,
            # because rename will replace the file anyway (which should be
            # identical).
            pathlib.PosixPath(file.fd.name).rename(file_path)
            return True
        else:
            pathlib.PosixPath(file.fd.name).unlink()
            return False

    def exists(self, filename: str) -> bool:
        """See FileCacherBackend.exists()."""
        file_path: pathlib.Path = self.path / filename

        return file_path.is_file()

    def describe(self, filename: str) -> str:
        """See FileCacherBackend.describe()."""
        file_path: pathlib.Path = self.path / filename

        if not file_path.is_file():
            raise KeyError('File not found.')

        return ''

    def get_size(self, filename: str) -> int:
        """See FileCacherBackend.get_size()."""
        file_path: pathlib.Path = self.path / filename

        if not file_path.is_file():
            raise KeyError('File not found.')

        return file_path.stat().st_size

    def delete(self, filename: str):
        """See FileCacherBackend.delete()."""
        file_path: pathlib.Path = self.path / filename

        file_path.unlink(missing_ok=True)

    def list(self) -> List[FileWithDescription]:
        """See FileCacherBackend.list()."""
        res = []
        for path in self.path.glob('*'):
            if path.is_file():
                res.append(
                    FileWithDescription(
                        filename=str(path.relative_to(self.path)), description=''
                    )
                )
        return res

    def path_for_symlink(self, filename: str) -> Optional[pathlib.Path]:
        file_path = self.path / filename
        if not file_path.is_file():
            raise KeyError('File not found.')
        return file_path
